import random

import nrgten.parser as parser
import pyvcon.vcontacts_wrapper as vcontacts_wrapper
import os
import numpy as np
import time


def get_macromol_list(pdb_file, atypes_dict, massdefs, verbose=False, skipmass_threshold=2,
                      cov_bond_length=2, ignore_hetatms=False, solve=False, unique_id=None,
                      use_precomp_vcon=False):
    with open(pdb_file) as f:
        lines = f.readlines()
    multimodel_flag = False
    for line in lines:
        if line[:4] == "ATOM":
            break
        if line.strip()[:5] == "MODEL":
            multimodel_flag = True
            break

    # This is the single model (as in X-ray files) case
    if not multimodel_flag:
        return [Macromol(lines, atypes_dict, massdefs, verbose=verbose,
                         skipmass_threshold=skipmass_threshold, cov_bond_length=cov_bond_length,
                         ignore_hetatms=ignore_hetatms, solve=solve, pdb_file=pdb_file, unique_id=unique_id,
                         use_precomp_vcon=use_precomp_vcon)]

    # This is the multi model (as in NMR files) case
    macromols = []
    start_i, stop_i = None, None
    count = 0
    for i, line in enumerate(lines):
        if line.strip()[:5] == "MODEL":
            assert stop_i is None
            start_i = i
        if line.strip()[:6] == "ENDMDL":
            assert start_i is not None
            stop_i = i
            if unique_id is None:
                mol_id = None
            else:
                mol_id = str(unique_id) + str(count)
            macromols.append(Macromol(lines[start_i:stop_i], atypes_dict, massdefs, verbose=verbose,
                                      skipmass_threshold=skipmass_threshold, cov_bond_length=cov_bond_length,
                                      ignore_hetatms=ignore_hetatms, solve=solve, pdb_file=pdb_file, unique_id=mol_id,
                                      use_precomp_vcon=use_precomp_vcon))
            count += 1
            start_i, stop_i = None, None
    assert start_i is None and stop_i is None
    return macromols

class Macromol:
    """ Class use to represent macromolecules. Infers the connectivity of
        residues from the atomic positions so that residue numbers and chain
        names have no effect on the data structure itself.
    """

    def __init__(self, pdb_lines, atypes_dict, massdefs, verbose=False, skipmass_threshold=2,
                 cov_bond_length=2, ignore_hetatms=False, solve=False, pdb_file=None, unique_id=None,
                 use_precomp_vcon=False):
        self.pdb_file = pdb_file
        self.unique_id = unique_id
        self.use_precomp_vcon = use_precomp_vcon
        self.dirpath = os.path.dirname(os.path.abspath(__file__))
        assert len(self.dirpath) > 0
        self.verbose = verbose
        self.skipmass_threshold = skipmass_threshold
        self.cov_bond_length = cov_bond_length
        self.atoms, self.resis, self.alt_flag = parser.parse_pdb_from_lines(pdb_lines, atypes_dict,
                                                                            ignore_hetatms=ignore_hetatms)
        self.index_atoms()
        if self.alt_flag:
            self.clear_alt_tags()
        self.atypes_dict, self.massdefs = atypes_dict, massdefs
        self.masses = self.compute_masses_general(self.atypes_dict, self.massdefs)
        self.masscoords = self.make_masscoords()
        self.solved = False
        self.coords_range, self.distmat, self.resi_cubes = None, None, None
        self.connect, self.bends, self.torsions = None, None, None
        self.interact_pairs_level_dict = None

        if solve:
            self.solve()

    def solve(self):
        self.distmat = self.make_distmat()
        self.resi_cubes = self.make_resi_cubes()
        self.print_verbose(", ".join([r for r in self.resis]))
        self.connect = self.compute_connect_general()
        self.bends = self.find_bends(self.connect)
        self.torsions = self.find_torsions(self.connect)
        self.interact_pairs_level_dict = self.find_interacting_pairs(self.torsions, self.bends, self.connect)
        self.solved = True

    def clear_info(self):
        self.distmat, self.resi_cubes, self.connect, self.interact_pairs = None, None, None, None
        self.solved = False

    def get_unique_id(self):
        """ Returns the unique id for that macromol object (used for multithreading).
        """
        if self.unique_id is None:
            raise ValueError("No unique_id given for macromolecule from file {0}".format(self.pdb_file))
        return self.unique_id

    def update(self):
        """ Updates the x, y, z coordinates of masses to reflect those of their respective center atoms.
        """
        new_masses = []
        for m in self.masses:
            center_atom = self.resis[m[4]][m[3]]
            x, y, z = center_atom.xyz
            new_masses.append((x, y, z, m[3], m[4], m[5]))
        self.masses = new_masses
        self.masscoords = self.make_masscoords()

    def index_atoms(self):
        for i, a in enumerate(self.atoms):
            a.set_macromol_index(i)

    def make_masscoords(self):
        masscoords = np.zeros((len(self.masses), 3))
        for i, m in enumerate(self.masses):
            masscoords[i] = np.array(m[0:3])
        return masscoords

    def make_distmat(self):
        """ Returns the mass distance matrix.
        """
        n = len(self.masses)
        distmat = np.zeros((n, n))
        for i in range(n):
            for j in range(i + 1, n):
                distmat[i][j] = distmat[j][i] = np.linalg.norm(self.masscoords[i] - self.masscoords[j])
        return distmat

    def make_resi_cubes(self):
        n = len(self.masses)
        inf = float('inf')
        mininf = float('-inf')
        resi_cubes = np.zeros((n, 6))
        delta = self.cov_bond_length / 2 + 0.01
        for i, r in enumerate(self.resis):
            resi = self.resis[r]
            x1, x2, y1, y2, z1, z2 = inf, mininf, inf, mininf, inf, mininf
            for a in resi:
                x, y, z = resi[a].xyz
                if x < x1:
                    x1 = x
                if x > x2:
                    x2 = x
                if y < y1:
                    y1 = y
                if y > y2:
                    y2 = y
                if z < z1:
                    z1 = z
                if z > z2:
                    z2 = z
            resi_cubes[i] = np.array([x1 - delta, x2 + delta, y1 - delta, y2 + delta, z1 - delta, z2 + delta])
        return resi_cubes

    def compute_masses_general(self, atypes_dict, massdefs):
        """ Computes a list of (x, y, z, center, residue, massdef) tuples, and
            computes the connectivity matrix.
        """
        massdefs_set = set()
        for r in massdefs:
            massdefs_set.add(massdefs[r])
        resi_redefinition = False
        for md in massdefs_set:
            if md.redefine_resis is not None:
                resi_redefinition = True
        if resi_redefinition:
            self.redefine_resis(atypes_dict, massdefs)
        masses = []
        newresis = {}
        for r in self.resis.keys():
            resname = r.split('|')[0]
            rd = self.resis[r]
            md = massdefs[resname]
            if md.n_masses == 1:
                c = md.centers[0]
                if c in rd:
                    a = rd[c]
                    k = self.make_mass_resikey(r, md.mass_names[0])
                    x, y, z = a.xyz
                    masses.append((x, y, z, c, k, md))
                    assert k not in newresis
                    newresis[k] = dict()
                    for atom in rd:
                        newresis[k][atom] = rd[atom]
                # TODO : find way of dealing with crappy residue portions (maybe modeRNA after all?)
                else:
                    raise ValueError("could not find center atom {0} in residue {1}, file {2}\n".format(
                                                                                                 c, r, self.pdb_file))
            else:
                for i in range(len(md.centers)):
                    (atoms, c, name) = md.get_atoms_center_name(i)
                    atoms = atoms.copy()
                    if c not in atoms:
                        print(r)
                        print(atoms)
                        continue
                    atoms.remove(c)
                    newresi = {}
                    for a in atoms:
                        if a in rd:
                            newresi[a] = rd[a]
                    if c in rd:
                        a = rd[c]
                        newresi[c] = a
                        k = self.make_mass_resikey(r, name)
                        x, y, z = a.xyz
                        masses.append((x, y, z, c, k, md))
                        assert k not in newresis
                        newresis[k] = newresi
                    elif len(newresi) <= self.skipmass_threshold:
                        continue
                    else:
                        raise ValueError("could not find center atom {0} in residue {1}\n".format(c, r))
        self.resis = newresis
        # TODO: create self.oldresis and write method to map unassigned atoms to the closest mass center
        return masses

    def compute_connect_general(self):
        """ Computes the connectivity matrix and allows for different types of
            masses. Distance between atoms of <= 2 A is used for covalent bond
            definition.
        """
        n = len(self.masses)
        connect = np.zeros((n, n))
        for i in range(n):
            for j in range(i + 1, n):
                if self.check_connect(i, j):
                    connect[i][j] = connect[j][i] = 1
        return connect

    def check_connect(self, i, j):
        if not self.overlapped_cubes(i, j):
            return False
        m1 = self.masses[i]
        m2 = self.masses[j]
        if m1[5] is not m2[5]:  # Check if mass definition is the same (otherwise, covalent connection is impossible)
            return False
        resi1 = self.resis[m1[4]]
        resi2 = self.resis[m2[4]]
        for key1 in resi1:
            atom1 = resi1[key1]
            for key2 in resi2:
                atom2 = resi2[key2]
                if np.linalg.norm(atom1.xyz - atom2.xyz) <= 2:
                    return True
        return False

    def overlapped_cubes(self, i, j):
        c1 = self.resi_cubes[i]
        c2 = self.resi_cubes[j]
        return not (c1[1] < c2[0] or c2[1] < c1[0] or c1[5] < c2[4] or c2[5] < c1[4] or c1[3] < c2[2] or c2[3] < c1[2])

    @staticmethod
    def make_mass_resikey(ogkey, name):
        sk = ogkey.split('|')
        sk[0] = sk[0] + ".{0}".format(name)
        return '|'.join(sk)

    def redefine_resis(self, atypes_dict, massdefs):
        """ Applies the REDEFINE_RESIS rules found in mass definition files.
            Does two passes: one to remove the atoms transferred, one to add
            them to the residues.
        """
        n = len(self.resis)
        resikeys = [x for x in self.resis.keys()]
        to_add = dict()
        # check all i -> j pairs for potential transfers of atoms
        for i in range(n):
            resi1 = self.resis[resikeys[i]]
            resname1 = resikeys[i].split('|')[0]
            md1 = massdefs[resname1]
            for j in range(n):  # unidirectional so iterate over all j
                if j == i:
                    continue
                if resikeys[j] in to_add:
                    continue
                resi2 = self.resis[resikeys[j]]
                resname2 = resikeys[j].split('|')[0]
                md2 = massdefs[resname2]
                if md1 is md2 and md1.redefine_resis is not None:
                    from_atom = md1.connect_init[0]
                    to_atom = md1.connect_init[1]
                    if from_atom in resi1 and to_atom in resi2 and \
                                    np.linalg.norm(resi1[from_atom].xyz - resi2[to_atom].xyz) <= 2:
                        rr = md1.redefine_resis
                        resikey = None
                        from_resi = None
                        if rr[0] == 1:
                            resikey = resikeys[j]
                            from_resi = resi1
                        else:
                            resikey = resikeys[i]
                            from_resi = resi2
                        to_add[resikey] = []
                        for atom in rr[1]:
                            if not atom in from_resi:
                                raise ValueError("Atom {0} not in residue {1}".format(atom, resikeys[i]))
                            to_add[resikey].append(from_resi[atom])
                            del from_resi[atom]
            # check if transfer atoms present and untransferred
            if md1.redefine_resis is not None:
                for atom in md1.redefine_resis[1]:
                    if atom in resi1:
                        del resi1[atom]
        # put transferred atoms back inside respective residues
        for resi in to_add:
            for atom in to_add[resi]:
                self.resis[resi][atom.name] = atom

    def compute_n_chains(self):
        """ Uses the connectivity matrix to infer the number of chains present
            in the system.
        """
        visited, chains = self.visit_dfs()
        return len(chains)

    def compute_seqs_dfs(self):
        """ Uses the connectivity matrix to get the sequence(s) of masses in the molecule. Chains must be contiguous.
        """
        visited, chains = self.visit_dfs()
        if self.validate_visited(visited):
            seqs = [[]]
            seqs_i = 0
            for i in range(len(self.masses)):
                if i > 0 and visited[i-1] != visited[i]:
                    seqs.append([])
                    seqs_i += 1
                seqs[seqs_i].append(self.masses[i][4].split('|')[0])
            return seqs
        else:
            raise ValueError("Sequences not sequential in file " + self.pdb_file)

    @staticmethod
    def validate_visited(visited):
        """ Returns True if chains are contiguous, False otherwise.
        """
        seen = set()
        last_v = None
        for v in visited:
            if v != last_v:
                if v in seen:
                    return False
                seen.add(v)
            last_v = v
        return True

    def visit_dfs(self):
        """ Creates a list of 'colors' for all the masses in the system using DFS. Same 'color' means the masses are
            part of the same chain. Returns the list, and a set of 'colors' (cardinality = number of chains).
        """
        n = len(self.connect)
        visited = [0] * n
        color = 1
        for i in range(n):
            self.dfs(self.connect, visited, color, i)
            color += 1
        chains = set(visited)
        if 0 in chains:
            raise ValueError("Some nodes were not reached by compute_n_chains")
        return visited, chains

    def dfs(self, connect, visited, color, i):
        """ Simple depth-first search algorithm, used for determining number of
            separate chains.
        """
        for e in self.adjacent_edges(connect, i):
            if visited[e] == 0 or visited[e] > color:
                visited[e] = color
                self.dfs(connect, visited, color, e)

    @staticmethod
    def adjacent_edges(connect, i):
        """ Returns all edges/vertices adjacent to vertex i in connectivity
            matrix connect (undirected graph).
        """
        ae = []
        for j in range(len(connect)):
            if connect[i][j] != 0:
                ae.append(j)
        return ae

    def find_bends(self, connect):
        """ Finds all (i, j, k) triplets such that there is a covalent i-j-k
            connection.
        """
        bends = set()
        n = len(connect)
        for i in range(n):
            for j in self.get_connects(connect, i):
                for k in self.get_connects(connect, j):
                    bend = self.make_bend_tuple(i, j, k)
                    if bend is not None:
                        bends.add(bend)
        return self.trim_bends(bends)

    @staticmethod
    def trim_bends(bends):
        """ Gets rid of duplicate bending angles.
        """
        trimmed = set()
        for b in bends:
            if b[0] > b[2]:
                b = (b[2], b[1], b[0])
            trimmed.add(b)
        return trimmed

    @staticmethod
    def make_bend_tuple(i, j, k):
        s = {i, j, k}
        if len(s) != 3:
            return None
        return i, j, k

    def find_torsions(self, connect):
        torsions = set()
        n = len(connect)
        for i in range(n):
            for j in self.get_connects(connect, i):
                for k in self.get_connects(connect, j):
                    for l in self.get_connects(connect, k):
                        torsion = self.make_torsion_tuple(i, j, k, l)
                        if torsion is not None:
                            torsions.add(torsion)
        return self.trim_torsions(torsions)

    @staticmethod
    def trim_torsions(torsions):
        trimmed = set()
        for t in torsions:
            if t[0] > t[3]:
                t = (t[3], t[2], t[1], t[0])
            trimmed.add(t)
        return trimmed

    def make_torsion_tuple(self, i, j, k, l):
        s = {i, j, k, l}
        self.print_verbose(s)
        if len(s) != 4:
            return None
        return i, j, k, l

    @staticmethod
    def get_connects(connect, i):
        connects = []
        for (j, c) in enumerate(connect[i]):
            if c != 0:
                connects.append(j)
        return connects

    def find_interacting_pairs(self, torsions, bends, connect):
        ips_l1 = set()
        ips_l2 = set()
        ips_l3 = set()
        for t in list(torsions):
            for i in range(len(t)):
                for j in range(i + 1, len(t)):
                    if t[i] < t[j]:
                        ips_l3.add((t[i], t[j]))
                    else:
                        ips_l3.add((t[j], t[i]))
        for b in list(bends):
            for i in range(len(b)):
                for j in range(i+1, len(b)):
                    if b[i] < b[j]:
                        ips_l2.add((b[i], b[j]))
                    else:
                        ips_l2.add((b[j], b[i]))
        for i in range(len(connect)):
            for j in range(i + 1, len(connect)):
                if connect[i][j] != 0:
                    ips_l1.add((i, j))
        ips_l2 = ips_l2.union(ips_l1)
        ips_l3 = ips_l3.union(ips_l2)
        self.print_verbose(ips_l1)
        self.print_verbose(ips_l2)
        self.print_verbose(ips_l3)
        return {0: set(), 1: ips_l1, 2: ips_l2, 3: ips_l3}

    def is_disconnected(self, i, j, level=3):
        ips = self.interact_pairs_level_dict[level]
        if i < j:
            return not ((i, j) in ips)
        else:
            return not ((j, i) in ips)

    def print_connect(self):
        for row in self.connect:
            print(row)

    def is_equal(self, other):
        """
        Args:
            other: another Macromol instance
        Returns: boolean
        """
        assert isinstance(other, Macromol)
        if len(self.atoms) != len(other.atoms):
            return False
        for i in range(len(self.atoms)):
            if not self.atoms[i].is_equal(other.atoms[i]):
                return False
        return True

    @staticmethod
    def pretty_print_interactions(interlist, masses):
        s = ""
        for i in interlist:
            for x in i:
                s = s + masses[x][4] + ", "
            s = s[:-2] + "\n"
        print(s)

    def print_verbose(self, string):
        if self.verbose:
            print(string)

    def rotate(self, rotation_mat):
        """ Performs the rotation described by the rotation matrix on all atoms. Original centroid is preserved.
        """
        centroid = self.get_centroid()
        self.translate_xyz(-centroid)
        all_atom_coords = self.get_allatom_coords()
        rotated_coords = np.dot(all_atom_coords, np.transpose(rotation_mat))
        for atom, vec in zip(self.atoms, rotated_coords):
            atom.set_xyz(vec)
        self.translate_xyz(centroid)

    def get_allatom_coords(self):
        """ Returns Nx3 matrix containing coordinates of atoms, in order of increasing atom number.
        """
        coords = np.zeros((len(self.atoms), 3))
        for i, atom in enumerate(self.atoms):
            coords[i] = atom.xyz
        return coords

    def get_n_atoms(self):
        """ Returns the number of atoms in the macromolecule(s).
        """
        return len(self.atoms)

    def translate_3n_matrix(self, vector):
        """ Translation with full vector represented as Nx3 matrix.
            Can (and most probably will) mess up internal structure.
        """
        masses = self.masses
        resis = self.resis
        assert len(masses) == len(vector)
        for i, m in enumerate(masses):
            vec = vector[i]
            assert len(vec) == 3
            atoms = resis[m[4]]
            for a in atoms:
                atom = atoms[a]
                atom.translate(vector[i])

    def get_3n_vector(self):
        """ Returns the coordinates of the masses as 3n-dimensional (x,y,z, x,y,z, ...) vector.
        """
        vec = np.zeros((3*len(self.masscoords)))
        for i in range(len(self.masscoords)):
            vec[3*i], vec[3*i + 1], vec[3*i + 2] = self.masscoords[i]
        return vec

    def get_filtered_xyz(self, filter):
        """ Returns the xyz coordinates of the masses for which the center's name is in the filter set.
        """
        if filter is None:
            return self.get_xyz()
        return np.array([[a[0], a[1], a[2]] for a in self.masses if a[4].split('|')[0].split('.')[-1] in filter])

    def get_xyz(self):
        """ Returns all xyz coordinates of the masses.
        """
        return np.array([[a[0], a[1], a[2]] for a in self.masses])

    def get_filtered_3n_vector(self, filter):
        """ Same as get_filtered_xyz but returns 3n-dimensional vector.
        """
        coords = self.get_filtered_xyz(filter)
        return np.reshape(coords, (3*len(coords)))

    def get_filtered_n(self, filter):
        return len(self.get_filtered_xyz(filter))

    def translate_3n_vector(self, vector):
        """ Translation with full vector (x,y,z, x,y,z, ...).
            Can (and most probably will) mess up internal structure.
        """
        masses = self.masses
        resis = self.resis
        assert len(masses) * 3 == len(vector)
        for i, m in enumerate(masses):
            atoms = resis[m[4]]
            j = 3 * i
            for a in atoms:
                atom = atoms[a]
                atom.translate([vector[j], vector[j + 1], vector[j + 2]])

    def set_bfactors(self, vector):
        masses = self.masses
        resis = self.resis
        assert len(vector) == len(masses)
        for i, m in enumerate(masses):
            atoms = resis[m[4]]
            for a in atoms:
                atom = atoms[a]
                atom.bfact = vector[i]

    def set_occupancy(self, vector):
        masses = self.masses
        resis = self.resis
        assert len(vector) == len(masses)
        for i, m in enumerate(masses):
            atoms = resis[m[4]]
            for a in atoms:
                atom = atoms[a]
                atom.occ = vector[i]

    def translate_xyz(self, vector):
        """ Translation with (x, y, z) vector. Preserves internal structure.
        """
        for atom in self.atoms:
            atom.translate(vector)

    def get_centroid(self):
        sum = np.array([0, 0, 0])
        for i in range(len(self.masses)):
            sum = np.add(sum, np.array(self.masses[i][0:3]))
        return sum / len(self.masses)

    def clear_alt_tags(self):
        for a in self.atoms:
            a.alt = ' '

    def write_to_file(self, filename):
        with open(filename, "w") as fh:
            self.write_to_filehandle(fh)

    def write_to_filehandle(self, fh):
        s = self.get_pdb_file_as_string()
        fh.write(s)

    def get_pdb_file_as_string(self):
        #
        # TODO : right now in the case of RNA, not all original atoms are part of resis. Find a workaround.
        s = ""
        resis = self.resis
        for resi in resis:
            atoms = sort_atoms(resis[resi])
            for atom in atoms:
                s += atom.toString() + "\n"
        return s + "END\n"

    def get_surface_dict(self):
        if self.alt_flag:
            self.clear_alt_tags()
        if self.use_precomp_vcon:
            tmp_pdb_file = self.pdb_file + ".tmp"
        else:
            tmp_pdb_file = self.pdb_file + "{}_{}.tmp".format(os.getpid(), random.randint(0, 1000))
        if not (self.use_precomp_vcon and os.path.isfile(tmp_pdb_file)):
            with open(tmp_pdb_file, "w") as f:
                self.write_to_filehandle(f)
        time.sleep(random.uniform(0.05, 0.1))
        sd = vcontacts_wrapper.get_surface_dict(tmp_pdb_file, self.get_n_atoms(), use_precomputed=self.use_precomp_vcon)
        if not self.use_precomp_vcon:
            os.remove(tmp_pdb_file)
        return sd


def sort_atoms(resi):
    adict = dict()
    for a in resi:
        atom = resi[a]
        num = atom.num
        adict[num] = atom
    atoms = []
    for k in sorted(adict.keys()):
        atoms.append(adict[k])
    return atoms



def write_hessian_helper_V1V4_optim():
    axes = ["X", "Y", "Z"]
    print('i3 = 3*i')
    # print('i3 = 3*i')
    # print('i3 = 3*i')
    print('j3 = 3*j')
    # print('j3 = 3*j')
    # print('j3 = 3*j')
    for a in ["i", "j"]:
        for b in ["i", "j"]:
            if a == b:
                sign = 1
            else:
                sign = -1
            i_n = a
            j_n = b
            for a_n, a_str in zip([0, 1, 2], axes):
                for b_n, b_str in zip([0, 1, 2], axes):
                    if sign == -1:
                        print('hessian[{0}3 + {1}][{2}3 + {3}] += {6} * const * bcoord["{4}"] * bcoord["{5}"]'.format(
                            i_n, a_n, j_n, b_n, a_str, b_str, sign))
                    else:
                        print('hessian[{0}3 + {1}][{2}3 + {3}] += const * bcoord["{4}"] * bcoord["{5}"]'.format(
                            i_n, a_n, j_n, b_n, a_str, b_str, sign))


if __name__ == "__main__":
    write_hessian_helper_V1V4_optim()


from nrgten.enm import ENM
import numpy as np
import os
import pickle


class ANM(ENM):
    """Class implementing the ANM (Anisotropic Network Model).

    Has 3 parameters: cutoff, power dependence and spring constant. Spring constant just rescales fluctuations but
    doesn't change their relative values. Cutoff can be infinity (fully connected model). The original ANM paper is:
    `doi.org/10.1016/S0006-3495(01)76033-X <https://doi.org/10.1016/S0006-3495(01)76033-X>`_ and the paper on varying
    the power dependence is: `doi.org/10.1073/pnas.0902159106 <https://doi.org/10.1073/pnas.0902159106>`_

    Note:
        ANM extends the `ENM` class, to see inherited attributes look at the documentation for `ENM`.

    Attributes:
        cut (float): the distance cutoff for interaction between 2 masses.
        pd (float): the power dependence of the interaction.
        kr (float): the interaction constant (does not change the solution but scales the values).
    """
    def __init__(self, pdb_file, cut=float('inf'), power_dependence=0, kr=1, solve=True, use_pickle=False,
                 added_atypes=None, added_massdef=None, ignore_hetatms=False, atypes_list=None, massdef_list=None,
                 solve_mol=True, one_mass=False):
        """Constructor for the ANM class.

        Args:
            pdb_file (str): The PDB file from which to build the model.
            cut (float, optional): The interaction cutoff.
            power_dependence (float, optional): The power dependence. For consistency with published papers, a power
                                                dependence of 0 means that the distance will be squared in the Hessian.
            solve (bool, optional): If True, the Hessian matrix will be built and solved.
            use_pickle (bool, optional): If True, the ENM object will be solved only once and subsequently built from
                                         its pickled representation. Uses a lot of disk space for large systems.
            ignore_hetatms (bool, optional): Flag to ignore HETATM records in the PDB file.
            atypes_list (list, optional): If supplied, the default .atypes configuration files are ignored and these
                                          are the only ones read.
            massdef_list (list, optional): If supplied, the default .masses configuration files are ignored and these
                                           are the only ones read.
            solve_mol (bool, optional): If True, the underlying Macromol object will be solved, i.e. the connectivity
                                        of the residues will be inferred.
            one_mass (bool, optional): If True, nucleic acids will be built using only one mass per nucleotide instead
                                       of 3.
        """
        self.cut = cut
        self.pd = power_dependence
        self.kr = kr
        super().__init__(pdb_file, added_atypes=added_atypes, added_massdef=added_massdef, solve=solve,
                         use_pickle=use_pickle, ignore_hetatms=ignore_hetatms, atypes_list=atypes_list,
                         massdef_list=massdef_list, solve_mol=solve_mol, one_mass=one_mass)

    def build_hessian(self):
        """Builds the Hessian matrix.

        Returns:
            The Hessian matrix.
        """
        if not self.mol.solved:
            self.mol.solve()
        masscoords = self.mol.masscoords
        distmat = self.mol.distmat
        n = len(masscoords)
        hessian = np.zeros((3*n, 3*n))
        for i in range(n):
            for j in range(i+1, n):
                dist = distmat[i][j]
                if dist <= self.cut:
                    dist_pd = dist ** (2 + self.pd)

                    # diagonal of the off-diagonal 3x3 element and update diagonal of diagonal element
                    for k in range(3):
                        val = 2 * self.kr * (masscoords[j][k] - masscoords[i][k]) ** 2 / dist_pd
                        hessian[3 * i + k][3 * j + k] = -val
                        hessian[3 * i + k][3 * i + k] += val
                        hessian[3 * j + k][3 * j + k] += val

                    # off-diagonals of the off-diagonal 3x3 element and update off-diagonal of diagonal element
                    for (k, l) in ((0, 1), (0, 2), (1, 2)):
                        val = 2 * self.kr * (masscoords[j][k] - masscoords[i][k]) * \
                              (masscoords[j][l] - masscoords[i][l]) / dist_pd
                        hessian[3 * i + k][3 * j + l] = -1 * val
                        hessian[3 * i + l][3 * j + k] = -1 * val
                        hessian[3 * i + k][3 * i + l] += val
                        hessian[3 * j + k][3 * j + l] += val
        for i in range(3 * n):
            for j in range(i + 1, 3 * n):
                hessian[j][i] = hessian[i][j]
        return hessian

    def _get_pickle_file(self):
        param_string = "{0}_{1}_{2}".format(self.cut, self.pd, self.kr)
        if self.one_mass:
            return '.'.join(self.pdb_file.split('.')[:-1]) + ".ANM_{0}_1n.pickle".format(param_string)
        else:
            return '.'.join(self.pdb_file.split('.')[:-1]) + ".ANM_{0}.pickle".format(param_string)

    def build_from_pickle(self):
        """Builds an ANM object from its pickled state.

        Returns:
            True if successful, False otherwise.
        """
        pickle_file = self._get_pickle_file()
        if not os.path.isfile(pickle_file):
            return False
        try:
            with open(pickle_file, 'rb') as f:
                pickled_anm = pickle.load(f)
        except EOFError:
            return False
        if not (self.cut == pickled_anm.cut and
                self.pd == pickled_anm.pd and
                self.kr == pickled_anm.kr):
            return False
        if not self.is_equal(pickled_anm):
            return False
        self.__dict__.update(pickled_anm.__dict__)
        self._reconstitute()
        return True

    def pickle(self):
        """Builds a pickled state from the ANM object.
        """
        super()._clear_info()
        pickle_file = self._get_pickle_file()
        with open(pickle_file, 'wb') as f:
            pickle.dump(self, f)




import collections

accepted_from_to = set(['(n)', '(n+1)', '(n-1)'])


class MassDef:
    ''' Class used to represent the definition of masses in a given
        (macro)molecule.
    '''

    def __init__(self, massdef_file):
        self.connect = None
        self.connect_init = None
        self.redefine_resis = None
        self.n_masses = None
        self.centers = None
        self.mass_names = None
        self.mass_atom_sets = None
        self.name = None
        self.parse_massdef(massdef_file)
        if self.connect_init is None:
            self.connect_init = self.connect

    def get_atoms_center_name(self, i):
        assert self.centers is not None and self.mass_names is not None and self.mass_atom_sets is not None
        assert len(self.centers) == len(self.mass_names) and len(self.centers) == len(self.mass_atom_sets)
        assert i < len(self.centers) and i >= 0
        return (self.mass_atom_sets[i], self.centers[i], self.mass_names[i])

    def parse_massdef(self, massdef_file):
        with open(massdef_file) as f:
            lines = f.readlines()
        i = 0
        self.name = massdef_file.split('/')[-1].split('.')[0]
        while True:
            if i >= len(lines):
                break
            line = lines[i]
            ll = line.strip().split(':')
            tag = ll[0].strip()
            rest = None
            if len(ll) >= 2:
                rest = ll[1].strip()

            if tag == "CONNECT":
                assert self.connect is None
                rl = rest.split('->')
                self.connect = (rl[0].strip(), rl[1].strip())

            elif tag == "CONNECT_INIT":
                assert self.connect_init is None
                rl = rest.split('->')
                self.connect_init = (rl[0].strip(), rl[1].strip())

            elif tag == "REDEFINE_RESIS":
                assert self.redefine_resis is None
                if self.connect_init is None:
                    raise ValueError("Trying to redefine residues without a CONNECT_INIT line")
                rl = rest.split('|')
                from_to = rl[0].strip()
                ftl = from_to.split('->')
                self.validate_ftl(ftl)
                fr = self.format_from_to(ftl[0])
                to = self.format_from_to(ftl[1])
                resilist = [x.strip() for x in rl[1].strip().split(',')]
                direction = to - fr
                self.redefine_resis = (direction, resilist)

            elif tag == "N_MASSES":
                assert self.n_masses is None
                self.n_masses = int(rest)

            elif tag == "CENTER":
                assert self.n_masses == 1
                self.centers = [rest.strip()]

            elif tag == "NAME":
                assert self.n_masses == 1
                assert self.mass_names is None
                self.mass_names = [rest.strip()]

            elif tag[:5] == "MASS_":
                n = int(tag.split('_')[-1])
                if n == 1:
                    assert self.centers is None
                    assert self.mass_names is None
                    assert self.mass_atom_sets is None
                    self.centers = []
                    self.mass_names = []
                    self.mass_atom_sets = []
                assert len(self.centers) == n - 1
                (name, atoms, center) = self.parse_one_mass(lines, i)
                i += 3
                self.centers.append(center)
                self.mass_names.append(name)
                self.mass_atom_sets.append(set(atoms))
            i += 1

    def parse_one_mass(self, lines, i):
        assert len(lines) > i + 3
        three_lines = lines[i + 1:i + 4]
        (name, atoms, center) = (None, None, None)
        for line in three_lines:
            ll = line.strip().split(':')
            tag = ll[0].strip()
            rest = None
            if len(ll) >= 2:
                rest = ll[1].strip()

            if tag == "NAME":
                assert name is None
                name = rest.strip()

            elif tag == "ATOMS":
                assert atoms is None
                atoms = [x.strip() for x in rest.split(',')]

            elif tag == "CENTER":
                assert center is None
                center = rest.strip()
        return (name, atoms, center)

    def format_from_to(self, s):
        s = s.strip()[2:-1]
        if len(s) > 0:
            s = int(s)
        else:
            s = 0
        return s

    def validate_ftl(self, ftl):
        ''' Used to ensure that the from -> to nomenclature used for residue
            redifinition is valid.
        '''
        for x in ftl:
            x = x.strip()
            if x not in accepted_from_to:
                raise ValueError("Error in REDEFINE_RESIS line : {0} is not accepted".format(x))


def massdefs_dicts_are_same(mdd1, mdd2):
    assert isinstance(mdd1, collections.Mapping) and isinstance(mdd2, collections.Mapping)
    if mdd1.keys() != mdd2.keys():
        return False
    for k in mdd1.keys():
        if not massdefs_are_same(mdd1[k], mdd2[k]):
            return False
    return True

def massdefs_are_same(md1, md2):
    assert isinstance(md1, MassDef) and isinstance(md2, MassDef)
    if not (md1.connect == md2.connect and
            md1.connect_init == md2.connect_init and
            md1.redefine_resis == md2.redefine_resis and
            md1.n_masses == md2.n_masses and
            md1.centers == md2.centers and
            md1.mass_names == md2.mass_names and
            md1.mass_atom_sets == md2.mass_atom_sets):
        return False
    return True


import nrgten.eq_filler as eq_filler
import numpy as np
import sys
import os
import pickle
from nrgten.enm import ENM


class ENCoM(ENM):
    """Class implementing the latest version of ENCoM (Elastic Network Contact Model).

        It extends the ENM class, implementing its own build_hessian(), pickle() and build_from_pickle() methods. See
        ENM class for the attributes it inherits. For details about the ENCoM model, see the 2014 Frappier and
        Najmanovich paper: `doi.org/10.1371/journal.pcbi.1003569 <https://doi.org/10.1371/journal.pcbi.1003569>`_

        Note:
            This class also implements the STeM model (Generalized Spring Tensor Model, on which ENCoM is based). The
            original STeM paper can be found here:
            `doi.org/10.1186/1472-6807-10-S1-S3 <https://doi.org/10.1186/1472-6807-10-S1-S3>`_

        Attributes:
            dirpath (str): The absolute path to the directory containing the encom.py module.
            use_stem (bool): If True, the model reverts back to STeM.
            pd (float): The power dependency of the long-range interaction.
            kr (float): The weight of the first term (bond stretching) from the ENCoM potential.
            ktheta (float): The weight of the second term (angle bending) from the ENCoM potential.
            kphi (float): The weight of the third term (torsion angles) from the ENCoM potential.
            epsi (float): The weight of the fourth term (long-range interactions) from the ENCoM potential.
            bfact_params (bool): If True, the best parameters for b-factor prediction (according to the 2014 paper)
                                 are used
            ic (float): The interaction constant for favorable pairs of atom types.
            inter_mat (list): The interaction matrix for atom types, in a list of lists format where every first
                              element is None (to allow indexing starting at 1 instead of 0).
            V1_H (numpy.ndarray): The Hessian for the 1st potential term (bond stratching).
            V2_H (numpy.ndarray): The Hessian for the 2nd potential term (angle bending).
            V3_H (numpy.ndarray): The Hessian for the 3rd potential term (torsion angles).
            V4_H (numpy.ndarray): The Hessian for the 4th potential term (long-range interactions).
            bij (numpy.ndarray): Matrix of the Bij term in the ENCoM potential, which is the modulating factor for the
                                 long-range interaction (V4)
        """

    def __init__(self, pdb_file, kr=1000, ktheta=10000, kphi=10000, epsi=0.01, apply_epsi=False, interact_const=3,
                 null_interact_const=1,
                 power_dependenceV4=4, interact_mat=None, use_stem=False, kphi1=1, kphi3=0.5, bfact_params=False,
                 added_atypes=None, added_massdef=None, atypes_list=None, massdef_list=None, verbose=False, solve=True,
                 ignore_hetatms=False, use_pickle=False, solve_mol=True, one_mass=False, interact_level=3,
                 teruel2021_legacy=False, new_optimized_params=False, use_precomp_vcon=False):
        """Constructor for the ENCoM class.

        Args:
            pdb_file (str): The PDB file to read.
            kr (float, optional): The weight of the first term (bond stretching) from the ENCoM potential.
            ktheta (float, optional): The weight of the second term (angle bending) from the ENCoM potential.
            kphi (float, optional): The weight of the third term (torsion angles) from the ENCoM potential.
            epsi (float, optional): The weight of the fourth term (long-range interactions) from the ENCoM potential.
            apply_epsi (bool, optional): If True, the other constants are multiplied by epsi.
            interact_const (float, optional): The interaction constant for favorable pairs of atom types.
            power_dependenceV4 (float, optional): The power dependency of the long-range interaction.
            interact_mat (list, optional): The interaction matrix for atom types, in a list of lists format where every
                                           first element is None (to allow indexing starting at 1 instead of 0).
            use_stem (bool, optional): If True, the model reverts back to STeM (on which ENCoM is based).
            kphi1 (float, optional): Scales the kphi constant.
            kphi3 (float, optional): Scales the kphi constant.
            bfact_params (bool, optional): If True, the best parameters for b-factor prediction (according to the 2014
                                           paper) are used.
            added_atypes (list, optional): list of .atypes configuration files to add.
            added_massdef (list, optional): list of .masses configuration files to add.
            atypes_list (list, optional): If supplied, the default .atypes configuration files are ignored and these
                                          are the only ones read.
            massdef_list (list, optional): If supplied, the default .masses configuration files are ignored and these
                                           are the only ones read.
            verbose (bool, optional): Triggers the verbose mode.
            solve (bool, optional): If True, the Hessian matrix will be built and solved.
            use_pickle (bool, optional): If True, the ENM object will be solved only once and subsequently built from
                                         its pickled representation. Uses a lot of disk space for large systems.
            ignore_hetatms (bool, optional): Flag to ignore HETATM records in the PDB file.
            solve_mol (bool, optional): If True, the underlying Macromol object will be solved, i.e. the connectivity
                                        of the residues will be inferred.
            one_mass (bool, optional): If True, nucleic acids will be built using only one mass per nucleotide instead
                                       of 3.
            teruel2021_legacy (bool, optional): If True, the parameters will perfectly reproduce the vibrational entropy
                                                calculations in the Teruel et al. 2021 PLoS Computational Biology paper
        """
        self.dirpath = os.path.dirname(os.path.abspath(__file__))
        assert len(self.dirpath) > 0
        self.use_stem = use_stem
        self.pd = power_dependenceV4
        if self.use_stem:
            kr = 100
            ktheta = 20
            kphi = 2.75
            epsi = 0.36
            apply_epsi = True
        else:
            kphi *= (kphi1 / 2 + kphi3 * 9 / 2)

        # Setting parameters ###########################################################################################
        if apply_epsi:
            kr *= epsi
            ktheta *= epsi
            kphi *= epsi
        self.kr = kr
        self.ktheta = ktheta
        self.kphi = kphi
        self.epsi = epsi
        self.ic = interact_const
        self.nic = null_interact_const
        self.bfact_params = bfact_params
        self.interact_level = interact_level
        self.teruel2021_legacy = teruel2021_legacy
        self.use_precomp_vcon = use_precomp_vcon
        if teruel2021_legacy:
            self.epsi = 0.001
        if new_optimized_params:
            self.kr = 4000
            self.ktheta = 80000
            self.kphi = 20000
            self.epsi = 0.08
            self.ic = 80
            self.nic = 0.0125
        if self.bfact_params: # optimal parameters for b-factor correlation from Frappier and Najmanovich 2014
            self.kr = 1000
            self.ktheta = 100000
            self.kphi = 1000
            self.epsi = 100
            self.ic = 3
        if interact_mat is None:
            ic = self.ic
            nc = self.nic
            # Presence of None is to enable indexing using the 1-8 numbers from Sobolev et al

            interact_mat = [[None] * 9,
                            [None, ic, ic, ic, nc, ic, ic, ic, ic],
                            [None, ic, nc, ic, nc, ic, ic, ic, nc],
                            [None, ic, ic, nc, nc, ic, ic, nc, ic],
                            [None, nc, nc, nc, ic, ic, ic, ic, ic],
                            [None, ic, ic, ic, ic, ic, ic, ic, ic],
                            [None, ic, ic, ic, ic, ic, ic, ic, ic],
                            [None, ic, ic, nc, ic, ic, ic, nc, ic],
                            [None, ic, nc, ic, ic, ic, ic, ic, nc]]
            if teruel2021_legacy:
                ic = 3
                interact_mat = [[None] * 9,
                                [None, ic, ic, ic, 1, ic, ic, ic, ic],
                                [None, ic, 1, ic, 1, ic, ic, ic, 1],
                                [None, ic, ic, 1, 1, ic, ic, 1, ic],
                                [None, 1, 1, 1, ic, ic, ic, ic, ic],
                                [None, ic, ic, ic, ic, ic, ic, ic, ic],
                                [None, ic, ic, ic, ic, ic, ic, ic, ic],
                                [None, ic, ic, 1, ic, ic, ic, 1, ic],
                                [None, ic, 1, ic, 1, ic, ic, ic, 1]]
        self.inter_mat = interact_mat
        ################################################################################################################

        self.V1_H, self.V2_H, self.V3_H, self.V4_H, self.bij = None, None, None, None, None
        super().__init__(pdb_file, added_atypes=added_atypes, added_massdef=added_massdef, atypes_list=atypes_list,
                         massdef_list=massdef_list, verbose=verbose, solve=solve, ignore_hetatms=ignore_hetatms,
                         use_pickle=use_pickle, solve_mol=solve_mol, one_mass=one_mass,
                         teruel2021_legacy=teruel2021_legacy, use_precomp_vcon=use_precomp_vcon)

    def _get_pickle_file(self):
        # TODO : better logic
        pickle_file = self.pdb_file.split('.')[0]
        if self.use_stem:
            pickle_file += ".stem.pickle"
        elif self.bfact_params:
            pickle_file += ".encom_bfacts.pickle"
        elif self.one_mass:
            pickle_file += ".encom_1n.pickle"
        else:
            pickle_file += ".encom.pickle"
        return pickle_file

    def build_from_pickle(self):
        """Builds an ENCoM object from its pickled state.

        Returns:
            True if successful, False otherwise.
        """
        pickle_file = self._get_pickle_file()
        if not os.path.isfile(pickle_file):
            return False
        try:
            with open(pickle_file, 'rb') as f:
                pickled_enc = pickle.load(f)
        except EOFError:
            return False
        if not (self.kr == pickled_enc.kr and
                self.ktheta == pickled_enc.ktheta and
                self.kphi == pickled_enc.kphi and
                self.epsi == pickled_enc.epsi and
                self.inter_mat == pickled_enc.inter_mat and
                self.pd == pickled_enc.pd and
                self.use_stem == pickled_enc.use_stem):
            return False
        if not self.is_equal(pickled_enc):
            return False
        self.__dict__.update(pickled_enc.__dict__)
        self._reconstitute()
        return True

    def pickle(self):
        """Builds a pickled state from the ENCoM object.
        """
        self.V1_H, self.V2_H, self.V3_H, self.V4_H, self.bij = None, None, None, None, None
        super()._clear_info()
        pickle_file = self._get_pickle_file()
        with open(pickle_file, 'wb') as f:
            pickle.dump(self, f)

    def build_hessian(self):
        """Builds the Hessian matrix.

        Returns:
            The Hessian matrix.
        """
        if not self.mol.solved:
            self.mol.solve()
        masses = self.mol.masses
        connect = self.mol.connect
        bends = self.mol.bends
        torsions = self.mol.torsions
        resis = self.mol.resis
        self.V1_H = self._build_V1_hessian(masses, connect)
        self.V2_H = self._build_V2_hessian(masses, bends)
        self.V3_H = self._build_V3_hessian(masses, torsions)
        self.V4_H = self._build_V4_hessian(masses, resis)
        V_tot = self._sum_hessians()
        return V_tot

    def build_hybrid_hessian(self, other, solve=True):
        """Builds the Hessian matrix with V1, V2, V3 from self and V4 from other.
        """
        assert isinstance(other, ENCoM)
        if not self.mol.solved:
            self.mol.solve()
        self.eigfreqs = None
        self.sq_eigvecs = None
        masses = self.mol.masses
        connect = self.mol.connect
        bends = self.mol.bends
        torsions = self.mol.torsions
        if not other.mol.solved:
            other.mol.solve()
        resis = other.mol.resis
        other_masses = other.mol.masses
        if self.V1_H is None:
            self.V1_H = self._build_V1_hessian(masses, connect)
        if self.V2_H is None:
            self.V2_H = self._build_V2_hessian(masses, bends)
        if self.V3_H is None:
            self.V3_H = self._build_V3_hessian(masses, torsions)
        if other.V4_H is None:
            other.V4_H = other._build_V4_hessian(other_masses, resis)
        self.h = np.add(np.add(np.add(self.V1_H, self.V2_H), self.V3_H), other.V4_H)
        if solve:
            self.solve()

    def build_hybrid_hessian_bij(self, other, solve=True):
        assert isinstance(other, ENCoM)
        if not self.mol.solved:
            self.mol.solve()
        self.eigfreqs = None
        self.sq_eigvecs = None
        masses = self.mol.masses
        connect = self.mol.connect
        bends = self.mol.bends
        torsions = self.mol.torsions
        resis = self.mol.resis
        if not other.mol.solved:
            other.mol.solve()
        other_resis = other.mol.resis
        other_masses = other.mol.masses
        if self.V1_H is None:
            self.V1_H = self._build_V1_hessian(masses, connect)
        if self.V2_H is None:
            self.V2_H = self._build_V2_hessian(masses, bends)
        if self.V3_H is None:
            self.V3_H = self._build_V3_hessian(masses, torsions)
        if other.bij is None:
            other.bij = other._compute_Bij(other_masses, other_resis)
        temp_V4_H = self._build_V4_hessian(masses, resis, bij=other.bij)
        self.h = np.add(np.add(np.add(self.V1_H, self.V2_H), self.V3_H), temp_V4_H)
        if solve:
            self.solve()


    def _sum_hessians(self):
        return np.add(np.add(np.add(self.V1_H, self.V2_H), self.V3_H), self.V4_H)

    def _build_V1_hessian(self, masses, connect_mat):
        n = len(masses)
        hessian = np.zeros((3 * n, 3 * n))
        distmat = self.mol.distmat
        for i in range(n):
            for j in range(i + 1, n):
                if connect_mat[i][j] != 0:
                    dist = distmat[i][j]
                    dist_sq = dist ** 2

                    # diagonal of the off-diagonal 3x3 element and update diagonal of diagonal element
                    for k in range(3):
                        val = 2 * self.kr * (masses[j][k] - masses[i][k]) ** 2 / dist_sq
                        hessian[3 * i + k][3 * j + k] += -1 * val
                        hessian[3 * i + k][3 * i + k] += val
                        hessian[3 * j + k][3 * j + k] += val

                    # off-diagonals of the off-diagonal 3x3 element and update off-diagonal of diagonal element
                    for (k, l) in ((0, 1), (0, 2), (1, 2)):
                        val = 2 * self.kr * (masses[j][k] - masses[i][k]) * (masses[j][l] - masses[i][l]) / dist_sq
                        hessian[3 * i + k][3 * j + l] += -1 * val
                        hessian[3 * i + l][3 * j + k] += -1 * val
                        hessian[3 * i + k][3 * i + l] += val
                        hessian[3 * j + k][3 * j + l] += val
        for i in range(3 * n):
            for j in range(i + 1, 3 * n):
                hessian[j][i] = hessian[i][j]
        return hessian

    def _build_V2_hessian(self, masses, bends):
        n = len(masses)
        hessian = np.zeros((3 * n, 3 * n))
        axes = ["X", "Y", "Z"]
        indexes = ["i", "j", "k"]
        for (b_i, bend) in enumerate(bends):
            self._print_verbose(b_i)
            self._print_verbose(bend)
            (i, j, k) = bend
            p = np.array(masses[i][:3]) - np.array(masses[j][:3])
            norm_p = np.linalg.norm(p)
            q = np.array(masses[k][:3]) - np.array(masses[j][:3])
            norm_q = np.linalg.norm(q)
            prod_norms = norm_p * norm_q
            norm_poverq = norm_p / norm_q
            norm_qoverp = norm_q / norm_p
            prod_norms_sq = prod_norms ** 2
            dot = np.dot(p, q)
            G = dot / prod_norms
            const = 2 * self.ktheta / (1 - G ** 2)  # constant term for filling the hessian

            co = dict()  # dictionary of coordinates (Xi, Xj, Xk, Yi, Yj, ...)
            for ax in zip([0, 1, 2], axes):
                for index in zip([i, j, k], indexes):
                    co[ax[1] + index[1]] = masses[index[0]][ax[0]]
            dGd = dict()
            for ax in axes:
                dGd[ax + "i"] = ((co[ax + "k"] - co[ax + "j"]) * prod_norms - dot * norm_qoverp * (
                co[ax + "i"] - co[ax + "j"])) / prod_norms_sq
                dGd[ax + "k"] = ((co[ax + "i"] - co[ax + "j"]) * prod_norms - dot * norm_poverq * (
                co[ax + "k"] - co[ax + "j"])) / prod_norms_sq
                dGd[ax + "j"] = ((2 * co[ax + "j"] - co[ax + "i"] - co[ax + "k"]) * prod_norms - dot * norm_qoverp * (
                co[ax + "j"] - co[ax + "i"]) -
                                 dot * norm_poverq * (co[ax + "j"] - co[ax + "k"])) / prod_norms_sq
            doubles = [x for x in zip(bend, indexes)]
            self._hessian_helper_V2V3(hessian, doubles, dGd, const)
        return hessian

    def _hessian_helper_V1V4(self, hessian, doubles, bcoord, const):
        axes = ["X", "Y", "Z"]
        for a in range(len(doubles)):
            for b in range(len(doubles)):
                i_n = doubles[a][0]  # this could hold i or j
                i_str = doubles[a][1]
                j_n = doubles[b][0]
                j_str = doubles[b][1]
                sign = 0
                if i_n == j_n:
                    sign = 1
                else:
                    sign = -1
                for a_n, a_str in zip([0, 1, 2], axes):
                    for b_n, b_str in zip([0, 1, 2], axes):
                        hessian[3 * i_n + a_n][3 * j_n + b_n] += sign * const * bcoord[a_str] * bcoord[b_str]
        return hessian

    def _hessian_helper_V1V4_optim(self, hessian, i, j, bcoord, const):
        """ This code was generated with func_writer.py
        """
        i3 = 3 * i
        j3 = 3 * j
        hessian[i3 + 0][i3 + 0] += const * bcoord["X"] * bcoord["X"]
        hessian[i3 + 0][i3 + 1] += const * bcoord["X"] * bcoord["Y"]
        hessian[i3 + 0][i3 + 2] += const * bcoord["X"] * bcoord["Z"]
        hessian[i3 + 1][i3 + 0] += const * bcoord["Y"] * bcoord["X"]
        hessian[i3 + 1][i3 + 1] += const * bcoord["Y"] * bcoord["Y"]
        hessian[i3 + 1][i3 + 2] += const * bcoord["Y"] * bcoord["Z"]
        hessian[i3 + 2][i3 + 0] += const * bcoord["Z"] * bcoord["X"]
        hessian[i3 + 2][i3 + 1] += const * bcoord["Z"] * bcoord["Y"]
        hessian[i3 + 2][i3 + 2] += const * bcoord["Z"] * bcoord["Z"]
        hessian[i3 + 0][j3 + 0] += -1 * const * bcoord["X"] * bcoord["X"]
        hessian[i3 + 0][j3 + 1] += -1 * const * bcoord["X"] * bcoord["Y"]
        hessian[i3 + 0][j3 + 2] += -1 * const * bcoord["X"] * bcoord["Z"]
        hessian[i3 + 1][j3 + 0] += -1 * const * bcoord["Y"] * bcoord["X"]
        hessian[i3 + 1][j3 + 1] += -1 * const * bcoord["Y"] * bcoord["Y"]
        hessian[i3 + 1][j3 + 2] += -1 * const * bcoord["Y"] * bcoord["Z"]
        hessian[i3 + 2][j3 + 0] += -1 * const * bcoord["Z"] * bcoord["X"]
        hessian[i3 + 2][j3 + 1] += -1 * const * bcoord["Z"] * bcoord["Y"]
        hessian[i3 + 2][j3 + 2] += -1 * const * bcoord["Z"] * bcoord["Z"]
        hessian[j3 + 0][i3 + 0] += -1 * const * bcoord["X"] * bcoord["X"]
        hessian[j3 + 0][i3 + 1] += -1 * const * bcoord["X"] * bcoord["Y"]
        hessian[j3 + 0][i3 + 2] += -1 * const * bcoord["X"] * bcoord["Z"]
        hessian[j3 + 1][i3 + 0] += -1 * const * bcoord["Y"] * bcoord["X"]
        hessian[j3 + 1][i3 + 1] += -1 * const * bcoord["Y"] * bcoord["Y"]
        hessian[j3 + 1][i3 + 2] += -1 * const * bcoord["Y"] * bcoord["Z"]
        hessian[j3 + 2][i3 + 0] += -1 * const * bcoord["Z"] * bcoord["X"]
        hessian[j3 + 2][i3 + 1] += -1 * const * bcoord["Z"] * bcoord["Y"]
        hessian[j3 + 2][i3 + 2] += -1 * const * bcoord["Z"] * bcoord["Z"]
        hessian[j3 + 0][j3 + 0] += const * bcoord["X"] * bcoord["X"]
        hessian[j3 + 0][j3 + 1] += const * bcoord["X"] * bcoord["Y"]
        hessian[j3 + 0][j3 + 2] += const * bcoord["X"] * bcoord["Z"]
        hessian[j3 + 1][j3 + 0] += const * bcoord["Y"] * bcoord["X"]
        hessian[j3 + 1][j3 + 1] += const * bcoord["Y"] * bcoord["Y"]
        hessian[j3 + 1][j3 + 2] += const * bcoord["Y"] * bcoord["Z"]
        hessian[j3 + 2][j3 + 0] += const * bcoord["Z"] * bcoord["X"]
        hessian[j3 + 2][j3 + 1] += const * bcoord["Z"] * bcoord["Y"]
        hessian[j3 + 2][j3 + 2] += const * bcoord["Z"] * bcoord["Z"]

    def _hessian_helper_V2V3(self, hessian, doubles, dGd, const):
        axes = ["X", "Y", "Z"]
        for a in range(len(doubles)):
            for b in range(len(doubles)):  # a and b are combinations of indexes : ii, ij, ik, etc...
                i_n = doubles[a][0]  # this can hold i, j or k
                i_str = doubles[a][1]
                j_n = doubles[b][0]
                j_str = doubles[b][1]
                for a_n, a_str in zip([0, 1, 2], axes):  # axis index and str repr
                    for b_n, b_str in zip([0, 1, 2], axes):
                        hessian[3 * i_n + a_n][3 * j_n + b_n] += const * dGd[a_str + i_str] * dGd[b_str + j_str]
        return hessian

    def _build_V3_hessian(self, masses, torsions):
        n = len(masses)
        hessian = np.zeros((3 * n, 3 * n))
        axes = ["X", "Y", "Z"]
        indexes = ["i", "j", "k", "l"]
        for (t_i, torsion) in enumerate(torsions):
            self._print_verbose(t_i)
            self._print_verbose(torsion)
            (i, j, k, l) = torsion
            a = np.array(masses[j][:3]) - np.array(masses[i][:3])
            b = np.array(masses[k][:3]) - np.array(masses[j][:3])
            c = np.array(masses[l][:3]) - np.array(masses[k][:3])
            v1 = np.cross(a, b)
            v2 = np.cross(b, c)
            dotv1v2 = np.dot(v1, v2)
            norm_v1 = np.linalg.norm(v1)
            norm_v2 = np.linalg.norm(v2)
            prod_norms = norm_v1 * norm_v2
            G = np.dot(v1, v2) / (norm_v1 * norm_v2)
            const = 2 * self.kphi / (1 - G ** 2)

            co = dict()  # dictionary of coordinates (Xi, Xj, Xk, Yi, Yj, ...)
            for ax in zip([0, 1, 2], axes):
                for index in zip([i, j, k, l], indexes):
                    co[ax[1] + index[1]] = masses[index[0]][ax[0]]
            dv1d = eq_filler.get_dv1d(co)
            dv2d = eq_filler.get_dv2d(co)
            dnorm_v1d = eq_filler.get_dnorm_v1d(co)
            dnorm_v2d = eq_filler.get_dnorm_v2d(co)
            G_const = prod_norms ** 2  # (|v1|*|v2|)^2
            dGd = dict()
            for ax in axes:
                for ind in indexes:
                    key = ax + ind
                    dGd[key] = ((np.dot(dv1d[key], v2) + np.dot(dv2d[key], v1)) * prod_norms - dotv1v2 * (
                    dnorm_v1d[key] * norm_v2 + dnorm_v2d[key] * norm_v1)) / G_const
            doubles = [x for x in zip(torsion, indexes)]
            self._hessian_helper_V2V3(hessian, doubles, dGd, const)
        return hessian

    def _build_V4_hessian(self, masses, resis, bij=None):
        n = len(masses)
        if bij is None:
            bij = self._compute_Bij(masses, resis)
            self.bij = bij
        hessian = np.zeros((3 * n, 3 * n))
        axes = ["X", "Y", "Z"]
        indexes = ["i", "j"]
        for i in range(n):
            for j in range(i + 1, n):
                if self._is_long_range(i, j):
                    self._print_verbose((i, j))
                    distij = self.mol.distmat[i][j]
                    bcoord = dict()  # corresponds to bx, by and bz in STeM.m
                    for ax in zip([0, 1, 2], axes):
                        bcoord[ax[1]] = masses[i][ax[0]] - masses[j][ax[0]]
                    distij_pd = distij ** self.pd
                    if self.use_stem:
                        const = 120 * self.epsi / distij_pd
                    else:
                        const = 120 * (self.epsi + bij[i][j]) / distij_pd
                    doubles = [x for x in zip([i, j], indexes)]
                    # self.hessian_helper_V1V4(hessian, doubles, bcoord, const)
                    self._hessian_helper_V1V4_optim(hessian, i, j, bcoord, const)
        return hessian


    def _is_long_range(self, i, j):
        if i == j:
            return False
        return self.mol.is_disconnected(i, j, level=self.interact_level)


def _distance(a, b):
    return ((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2 + (a[2] - b[2]) ** 2) ** 0.5

if __name__ == "__main__":
    test = ENCoM("../../tests/test_medium.pdb")
    print(test.build_motion_elastic(7, "test_elastic.pdb"))



from nrgten.atom import Atom
from nrgten.residue_list import ResidueList


def parse_pdb(pdb_file, atypes_dict):
    with open(pdb_file) as f:
        lines = f.readlines()
    return parse_pdb_from_lines(lines, atypes_dict)


def parse_pdb_from_lines(lines, atypes_dict, ignore_hetatms=False):
    resis = ResidueList(atypes_dict)
    for line in lines:
        if line[:4] == 'ATOM' or (not ignore_hetatms and line[:6] == 'HETATM'):
            resis.add(Atom(line))
    return resis.get_ordered_atoms(), resis.get_resilist(), resis.alt_flag



from nrgten.encom import ENCoM
from nrgten.anm import ANM
import numpy as np
# import sys
# import glob
# import os
# from scipy.stats import pearsonr
# from multiprocessing import Pool, current_process


def get_transit_probs(enm_a, enm_b, gamma=1, alignment=None):
    """Computes the transition 'probabilities' of reaching state B from state A and vice versa.

    Args:
        enm_a (ENM): The ENM object representing state A.
        enm_b (ENM): The ENM object representing state B.
        gamma (float, optional): The Boltzmann scaling factor.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]

    Returns:
        tuple: tuple containing:
            p_a(float): the probability of reaching state A from state B

            p_b(float): the probability of reaching state B from state A
    """
    vals_a = enm_a.eigvals[6:]
    vals_b = enm_b.eigvals[6:]
    if alignment is None:
        assert len(vals_a) == len(vals_b)
    else:
        assert len(alignment[0]) == len(alignment[1])
        assert len(vals_a) >= len(alignment[0])*3-6 and len(vals_b) >= len(alignment[0])*3-6
    pjs_a = np.exp(-1 * np.array(vals_a) / gamma)
    pjs_b = np.exp(-1 * np.array(vals_b) / gamma)
    pjs_a /= np.sum(pjs_a)
    pjs_b /= np.sum(pjs_b)
    if alignment is None:
        overlaps_a = get_overlaps(enm_b, enm_a, len(vals_a))
        overlaps_b = get_overlaps(enm_a, enm_b, len(vals_a))
    else:
        overlaps_a = get_overlaps(enm_b, enm_a, len(alignment[0]) * 3 - 6, [alignment[1], alignment[0]])
        overlaps_b = get_overlaps(enm_a, enm_b, len(alignment[0]) * 3 - 6, alignment)
    n = len(overlaps_a)
    p_a = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_a, pjs_a[:n])]))
    p_b = np.sum(np.array([x[0] * x[1] for x in zip(overlaps_b, pjs_b[:n])]))
    return p_a, p_b


def fit(reference, target, alignment=None, filter=None):
    """Fits the target to the reference.

    Only the masses that have centers in the filter set are kept (all are kept if filter is None). An alignment can also
    be applied to select corresponding masses. It is applied after filtering.

    Note:
        Only the conformational change vector is returned, and the target ENM object is left untouched. This vector
        does not include the 6 trivial 3-D degrees of freedom (rotational and translational).

    Warning:
        The target object is manipulated during this operation. Could lead to nasty effects if used naively during
        parallel computing.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector representing the conformational change from reference to target
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter) # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    target._rotate(r)
    target._translate_xyz(transvec)

    # new target xyz after translation and rotation
    targ_xyz_before = target.get_filtered_xyz(filter)
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    conf_change = np.array(listify(np.subtract(targ_xyz, ref_xyz)))
    target._translate_xyz(-transvec)
    target._rotate(np.transpose(r))
    return conf_change


def get_amplitudes_for_fit(reference, target, n_modes, alignment=None, filter=None):
    """This gives the amplitudes to apply to each eigenvector of the target to get as close as possible to the target.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which coefficients will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        np.ndarray: a vector of amplitudes of length `n_modes`
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = reference.get_filtered_eigvecs_mat(alignment[0], filter)
    x = np.dot(np.transpose(eigvecs), conf_change)
    return x[6:n_modes+6]


def cumulative_overlap(reference, target, n_modes, alignment=None, filter=None):
    """Cumulative overlap, as defined in Zimmerman and Jernigan 2014, Eq. 4.

    This computes the cumulative overlap between the `n_modes` first normal modes of the reference and the
    conformational change going from reference to target. See Eq. 4 from
    `doi.org/10.1261/rna.041269.113 <https://doi.org/10.1261/rna.041269.113>`_

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the cumulative overlap will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        float: the cumulative overlap, using the `n_modes` first normal modes
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter, n_vecs=6+n_modes))
    assert len(conf_change) == len(eigvecs[0])
    return _cumulative_overlap_helper(conf_change, eigvecs[6:6+n_modes])


def get_overlaps(reference, target, n_modes, alignment=None, filter=None):
    """Computes the overlaps between the `n_modes` from target and the conformational change.

    The conformational change is the vector of displacement going from `reference` to `target`, without
    translational/rotational degrees of freedom.

    Args:
        reference (ENM): The reference ENM object to which the target will be fitted.
        target (ENM): The target ENM object which will be fitted to the reference.
        n_modes (int): The number of normal modes for which the overlaps will be returned.
        alignment (list, optional): list of 2 lists of matching indices. Example: The alignment of sequences
                                    enm_a: -XYY-Z and enm_b: ZXYYXZ would give alignment=[[0,1,2,3],[1,2,3,5]]
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Returns:
        list: The list of overlaps, of length `n_modes`.
    """
    if alignment is None:
        assert reference.get_n_masses() == target.get_n_masses()
        n = reference.get_n_masses()
        alignment = [[x for x in range(n)], [x for x in range(n)]]
    conf_change = fit(reference, target, alignment, filter)
    eigvecs = np.transpose(reference.get_filtered_eigvecs_mat(alignment[0], filter))
    assert len(conf_change) == len(eigvecs[0])
    overlaps = []
    n = n_modes + 6
    if n > len(eigvecs):
        n = len(eigvecs)
    for i in range(6, n):
        vec = np.array(eigvecs[i])
        overlaps.append(np.abs(np.dot(vec, conf_change))/(np.linalg.norm(vec)*np.linalg.norm(conf_change)))
    return overlaps


def rmsd_alignment(reference, target, alignment, filter):
    assert len(alignment) == 2 and len(alignment[0]) == len(alignment[1])

    ref_xyz_before = reference.get_filtered_xyz(filter)  # before applying alignment
    targ_xyz_before = target.get_filtered_xyz(filter)
    n = len(alignment[0])
    ref_xyz = np.zeros((n, 3))
    targ_xyz = np.zeros((n, 3))
    for i in range(n):
        ref_xyz[i] = ref_xyz_before[alignment[0][i]]
        targ_xyz[i] = targ_xyz_before[alignment[1][i]]
    rmsd, r, transvec = rmsd_kabsch(ref_xyz, targ_xyz)
    return rmsd


def pca_ensemble(enm, macromols_list=None, variance_to_explain=0.99, filter=None):
    """Principal component analysis on an ensemble of structures.

    Args:
        enm (ENM): An ENM object from which the starting structure will be taken. If the macromols_list optional
                   argument is None, this object must also contain at least one other state of the macromolecule (the
                   PDB file must be in NMR format with MODEL and ENDMDL records).
        macromols_list (list, optional): A list of Macromol objects which will be used to compute the PCs.
        variance_to_explain (float, optional): The target proportion of variance explained by the PCs.
        filter (set, optional): set containing the names (str) of the center atoms for the masses to keep.

    Warning:
        The structures other than the starting structures are restored to their initial states at the end but are
        temporarily disturbed during the computation. This could lead to nasty effects if parallel computing is
        attempted without care.

    Returns:
        tuple: tuple containing:
            proportion_variance(numpy.ndarray): the proportion of variance explained by each PC

            components(numpy.ndarray): the PCs in matrix notation, 1 PC per row
    """
    if macromols_list is None:
        macromols_list = enm.mols[1:]
    refmol = enm.mol
    refmol_coords = refmol.get_3n_vector()
    ref_mean = get_mean_xyz(refmol_coords)
    refmol.translate_xyz(ref_mean)

    if filter is None:
        coords = np.zeros((3*len(refmol.masses), len(macromols_list)+1))
    else:
        raise ValueError("filtering not yet supported for PCA ensembles")
        coords = np.zeros((3*refmol.get_filtered_n(filter), len(macromols_list) + 1))
    for i, mol in enumerate([refmol] + macromols_list):
        transvec, r = None, None
        if i > 0:
            if not mol.solved:
                mol.solve()
            rmsd, r, transvec = _fit_to_reference_macromol_DEPRECATED(refmol, mol)
        if filter is None:
            coords[:,i] = mol.get_3n_vector()
        else:
            raise ValueError("filtering not yet supported for PCA ensembles")
            coords[:, i] = mol.get_filtered_3n_vector(filter)
        if i > 0:
            mol.translate_xyz(-transvec)
            mol.rotate(np.transpose(r))
    # Centering the data
    means = np.zeros((len(coords)))
    for i in range(len(coords)):
        means[i] = np.mean(coords[i])
        coords[i] -= means[i]
    covmat = np.cov(coords)
    eigvals, eigvecsmat = np.linalg.eigh(covmat)
    eigvecs = np.transpose(eigvecsmat)
    s = np.sum(eigvals)
    proportion_variance = eigvals/s
    if variance_to_explain == "all":
        return np.flip(proportion_variance, axis=0), np.flip(eigvecs, axis=0)
    total_var = 0
    n_significant = None
    for i in range(len(proportion_variance)-1, -1, -1):
        total_var += proportion_variance[i]
        if total_var > variance_to_explain:
            n_significant = len(proportion_variance) - i
            break
    return np.flip(proportion_variance, axis=0)[0:n_significant], np.flip(eigvecs, axis=0)[0:n_significant]


def get_mean_xyz(vector_3n):
    assert len(vector_3n) % 3 == 0
    mean_xyz = np.zeros(3)
    for offset in [0, 1, 2]:
        for i in range(0, len(vector_3n), 3):
            mean_xyz[offset] += vector_3n[i+offset]
    mean_xyz /= (len(vector_3n)/3)
    return mean_xyz


def get_pcs_no_rot_tran(enm_model, proportion_nrt_variance=0.99):
    assert len(enm_model.mols) > 1
    if enm_model.eigvals is None:
        enm_model.solve()
    variances, pcs = pca_ensemble(enm_model, variance_to_explain="all")
    rot_tran_vecs = enm_model.eigvecs[:6]
    nrt_vars = np.zeros(len(variances))
    nrt_vars_sum = np.zeros(len(variances))
    vals = np.zeros(len(variances))
    for i, pc in enumerate(pcs):
        val = rmsip([pc], rot_tran_vecs)
        # val = rmsip(rot_tran_vecs, [pc])
        vals[i] = val
        nrt_vars[i] = variances[i] * (1 - val)
    nrt_vars /= np.sum(nrt_vars)
    nrt_vars, pcs = reorder_pcs(pcs, nrt_vars)

    n_vecs = None
    for i in range(len(nrt_vars)):
        nrt_vars_sum[i] = np.sum(nrt_vars[:i+1])
        if nrt_vars_sum[i] >= proportion_nrt_variance:
            n_vecs = i+1
            break
    if n_vecs is None:
        raise ValueError("Not able to get to target variance proportion")
    max_vecs = len(pcs)
    n_target_vecs = n_vecs+6
    if n_target_vecs > max_vecs:
        n_target_vecs = max_vecs
    return nrt_vars[:n_vecs], _gram_schmidt(np.append(rot_tran_vecs, pcs, axis=0), n_target_vecs)[6:]


def reorder_pcs(pcs, nrt_vars):
    all_array = np.zeros((len(pcs), len(pcs[0])+1))
    all_array[:, 0] = nrt_vars
    all_array[:, 1:] = pcs
    sorted_array = all_array[np.argsort(-all_array[:, 0])]
    return sorted_array[:, 0], sorted_array[:, 1:]


def _gram_schmidt(vectors, target_n):
    # TODO : apply this to RMSIP to get rid of rot/trans contibution
    """ From https://gist.github.com/iizukak/1287876/edad3c337844fac34f7e56ec09f9cb27d4907cc7#gistcomment-1871542.
        Gram-schmidt orthonormalization of row vectors.
    """
    basis = []
    for v in vectors:
        x = np.zeros((len(v)))
        for b in basis:
            x += np.dot(v, b)*b
        w = v - x
        # w = v - np.sum(np.dot(v,b)*b  for b in basis) # Calling np.sum(generator) is deprecated
        if (w > 1e-10).any():
            basis.append(w/np.linalg.norm(w))
        if len(basis) == target_n:
            break
    if len(basis) < target_n:
        raise ValueError("Unable to produce {0} orthonormalized vectors".format(target_n))
    return np.array(basis)


def rmsip(ref_vectors, vectors_list):
    """Root mean square inner product, as defined in Leo-Macias et al. 2005.

    Link to the paper: `doi.org/10.1529/biophysj.104.052449 <https://doi.org/10.1529/biophysj.104.052449>`_

    Args:
        ref_vectors (np.ndarray): the reference vectors (PCs), in matrix form with one vector per row.
        vectors_list (np.ndarray): the vectors to test (eigenvectors describing normal modes) in matrix form with one
                                   vector per row.
    Returns:
        float: The root mean square inner product (RMSIP).
    """
    s = 0
    for r in ref_vectors:
        for v in vectors_list:
            s += np.square(np.dot(r, v))
    return np.sqrt(s/len(ref_vectors))


def nco(ref_vectors, ref_variances, vectors_list):
    ref_norms = np.zeros(len(ref_vectors))
    norms_list = np.zeros(len(vectors_list))
    for j, vec in enumerate(vectors_list):
        norms_list[j] = np.linalg.norm(vec)
    nco_result = []
    for i, ref_vec in enumerate(ref_vectors):
        nco_result.append([])
        ref_norms[i] = np.linalg.norm(ref_vec)
        overlaps_sq = np.zeros(len(vectors_list))
        for j, vec in enumerate(vectors_list):
            overlaps_sq[j] = (np.dot(vec, ref_vec)/(norms_list[j]*ref_norms[i]))**2
            nco_result[i].append(ref_variances[i]*np.sqrt(np.sum(overlaps_sq[:j+1])))
    return nco_result


def _fit_to_reference_macromol_DEPRECATED(ref, target):
    rmsd, r, transvec = rmsd_kabsch(ref.masscoords, target.masscoords)
    target.rotate(r)
    target.translate_xyz(transvec)
    return rmsd, r, transvec


def _cumulative_overlap_helper(ref_vector, vectors_list):
    s = 0
    ref_norm = np.linalg.norm(ref_vector)
    for vec in vectors_list:
        s += np.square(np.absolute(np.dot(ref_vector, vec))/(np.linalg.norm(vec)*ref_norm))
    return np.sqrt(s)



def _write_model(enm, count, fh):
    """ Writes one model to the given filehandle, with model number count.
    """
    fh.write("MODEL     {:>4}\n".format(count))
    enm.write_to_filehandle(fh)
    fh.write("ENDMDL\n")


def _motion(enm, mode_index, stepsize, maxrmsd, filename):
    """ Writes a PDB file displaying the motion of a normal mode, using MODEL records.
        Important note: mode_index starts at 1, which is the first nontrivial motion (7th normal mode).
    """
    mode = np.copy(enm.eigvecs[mode_index+5])
    _motion_3n_vector(enm, mode, stepsize, maxrmsd, filename)


def _motion_3n_vector(enm, vector_3n, stepsize, maxrmsd, filename):
    modermsd = _rmsd_of_3n_vector(vector_3n)
    vector_3n *= stepsize / modermsd
    nsteps = int(maxrmsd / stepsize)
    count = 1
    offset = 0
    with open(filename, "w") as fh:
        _write_model(enm, count, fh)
        for i, sign in enumerate([1, -1, 1, -1]):
            vector_3n *= sign
            if i == 3:
                offset = 1
            for i in range(nsteps - offset):
                enm._translate_3n_vector(vector_3n)
                _write_model(enm, count, fh)
                count += 1


def _rmsd_of_3n_vector(vec):
    dists = np.zeros((int(len(vec)/3)))
    for i in range(0, len(vec), 3):
        dists[int(i/3)] = vec[i]**2 + vec[i+1]**2 + vec[i+2]**2
    return np.sqrt(np.mean(dists))


def listify(matNx3):
    """ Takes an N x 3 matrix and makes it into a list (eigenvector-like).
    """
    l = []
    for row in matNx3:
        assert len(row) == 3
        for e in row:
            l.append(e)
    return l


def rmsd_kabsch(q, p):
    """ Kabsch algorithm for RMSD. Returns RMSD, rotation matrix and translation
        vector to optimally superpose p into q. See https://en.wikipedia.org/wiki/Kabsch_algorithm
    """
    assert len(p) == len(q)  # make sure that p and q are Nx3 matrices (x,y,z coordinates of N points)
    for i in range(len(p)):
        assert len(p[i]) == 3
        assert len(q[i]) == 3
    p = clone_nx3(p)
    q = clone_nx3(q)
    p_centroid = get_centroid(p)
    q_centroid = get_centroid(q)
    p_centered = subtract_centroid(p, p_centroid)
    q_centered = subtract_centroid(q, q_centroid)
    h = get_cross_covariance_matrix(p_centered, q_centered)
    u, s, vt = np.linalg.svd(h)
    v = np.transpose(vt)
    ut = np.transpose(u)
    d = np.linalg.det(np.dot(v, ut))
    x = np.zeros((3, 3))
    x[0][0] = 1
    x[1][1] = 1
    x[2][2] = d
    r = np.dot(np.dot(v, x), ut)
    rotated_centered_p = np.dot(p_centered, np.transpose(r))
    translation_vector = q_centroid - p_centroid
    rmsd = np.sqrt(np.mean(np.square(get_distance_vector(q_centered, rotated_centered_p))))
    return rmsd, r, translation_vector


def get_distance_vector(q, p):
    assert len(p) == len(q)
    d = np.zeros((len(p)))
    for i in range(len(p)):
        d[i] = np.linalg.norm(p[i]-q[i])
    return d


def get_cross_covariance_matrix(p, q):
    return np.dot(np.transpose(p), q)


def subtract_centroid(p, centroid):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j] - centroid[j]
    return new_p


def clone_nx3(p):
    new_p = np.zeros((len(p), 3))
    for i in range(len(p)):
        for j in range(3):
            new_p[i][j] = p[i][j]
    return new_p


def get_centroid(p):
    """ Returns the centroid of a set of point p (in the form of an Nx3 matrix).
    """
    n = len(p)
    xyz = np.array([0, 0, 0])
    for i in range(n):
        xyz = xyz + np.array([p[i][0], p[i][1], p[i][2]])
    return xyz/n




if __name__ == "__main__":
    # test_boltzmann_squares()


    open_state = ENCoM("../../tests/open_clean.pdb")
    closed_state = ENCoM("../../tests/closed_clean.pdb")
    print(get_amplitudes_for_fit(open_state, closed_state, 10))
    # print(overlap(open_state, closed_state, 10))
    # print(overlap(closed_state, open_state, 10))
    # print(open_state.compute_bfactors()[:15])




from nrgten.macromol import get_macromol_list
from nrgten.massdef import MassDef, massdefs_dicts_are_same
import numpy as np
import os
import abc
import sys


class ENM(metaclass=abc.ABCMeta):
    """Abstract base class (ABC) for Elastic Network Models (ENMs).

    The ENM class implements methods that are common for all implemented ENMs (ENCoM, ANM and STeM as of this writing)
    as well as the basic constructor. Subclasses are responsible for implementing the build_hessian() method (which
    generates the Hessian matrix) as well as the pickle() and build_from_pickle() methods which allow the pickling
    of solved ENMs for greatly accelerated subsequent computation of various metrics and/or properties.

    Attributes:
        dirpath (str): The absolute path to the directory containing the enm.py module.
        verbose (bool): If True, verbose mode is on.
        pickle_version (str): Ensures that ENMs pickled in the past are consistent with the current pickling scheme.
        one_mass (bool): For nucleic acids, False means 3 masses per nucleotide and True means 1 mass/nucleotide
        atypes_dict (dict): Dictionary of atom types built from the .atomtypes configuration files.
        massdefs (MassDef): Mass definitions built from the .masses configuration files.
        pdb_file (str): The PDB file on which the ENM is run.
        mols (list): list of Macromol objects representing the macromolecules in the PDB file (one per MODEL).
        mol (Macromol): The first Macromol (and the only in case of single-model PDBs), on which the ENM will be solved.
        not_there (set): Set of atoms that are not defined in the .atomtypes and .masses config files (normally empty).
        h (numpy.ndarray): The Hessian matrix.
        eigvecs (numpy.ndarray): The eigenvectors, in row format (one eigenvector per row).
        eigvals (list): list of eigenvalues, from lowest to highest
        eigvecs_mat (numpy.ndarrary): The eigenvectors in column format (one eigenvector per column).
        bfacts (list): The computed b-factors.
        entropy (float): The computed vibrational entropy.
    """
    def __init__(self, pdb_file, added_atypes=None, added_massdef=None, atypes_list=None, massdef_list=None,
                 verbose=False, solve=False, use_pickle=False, ignore_hetatms=False, solve_mol=True, one_mass=False,
                 teruel2021_legacy=False, use_precomp_vcon=False):
        """Constructor for the ENM abstract base class.

        Args:
            pdb_file (str): The PDB file to read.
            added_atypes (list, optional): list of .atypes configuration files to add.
            added_massdef (list, optional): list of .masses configuration files to add.
            atypes_list (list, optional): If supplied, the default .atypes configuration files are ignored and these
                                          are the only ones read.
            massdef_list (list, optional): If supplied, the default .masses configuration files are ignored and these
                                           are the only ones read.
            verbose (bool, optional): Triggers the verbose mode.
            solve (bool, optional): If True, the Hessian matrix will be built and solved.
            use_pickle (bool, optional): If True, the ENM object will be solved only once and subsequently built from
                                         its pickled representation. Uses a lot of disk space for large systems.
            ignore_hetatms (bool, optional): Flag to ignore HETATM records in the PDB file.
            solve_mol (bool, optional): If True, the underlying Macromol object will be solved, i.e. the connectivity
                                        of the residues will be inferred.
            one_mass (bool, optional): If True, nucleic acids will be built using only one mass per nucleotide instead
                                       of 3.
            teruel2021_legacy (bool, optional): If True, the parameters will perfectly reproduce the vibrational entropy
                                                calculations in the Teruel et al. 2021 PLoS Computational Biology paper
        """
        self.dirpath = os.path.dirname(os.path.abspath(__file__))
        assert len(self.dirpath) > 0
        self.verbose = verbose
        self.pickle_version = "no clearing of macromol"
        self.one_mass = one_mass
        if atypes_list is None:
            if self.one_mass:
                atomfiles = ["ribonucleic_acids_1n.atomtypes", "amino_acids.atomtypes"]
            else:
                atomfiles = ["ribonucleic_acids.atomtypes", "amino_acids.atomtypes"]
            atypes_list = [os.path.join(*[self.dirpath, "config", x]) for x in
                           atomfiles]
        if added_atypes is not None:
            atypes_list.append(added_atypes)
        self.atypes_dict = self._build_atypes_dict(atypes_list)
        if massdef_list is None:
            if self.one_mass:
                massfiles = ["ribonucleic_acids_1n.masses", "amino_acids.masses"]
            else:
                massfiles = ["ribonucleic_acids.masses", "amino_acids.masses"]
            massdef_list = [os.path.join(*[self.dirpath, "config", x]) for x in
                            massfiles]
        if added_massdef is not None:
            massdef_list.append(added_massdef)
        self.massdefs = self._parse_massdefs(atypes_list, massdef_list)
        self._print_verbose(self.massdefs)
        self.pdb_file = pdb_file
        self.teruel2021_legacy = teruel2021_legacy
        self.use_precomp_vcon=False
        self.mols = get_macromol_list(self.pdb_file, self.atypes_dict, self.massdefs, ignore_hetatms=ignore_hetatms,
                                      use_precomp_vcon=use_precomp_vcon)
        self.mol = self.mols[0]
        self.not_there = set()  # used to add resi-atom type pairs that are undefined
        self.h, self.eigvecs, self.eigvals, self.eigvecs_mat = None, None, None, None
        self.eigfreqs, self.sq_eigvecs, self.bfacts, self.entropy = None, None, None, None
        if solve:
            if use_pickle:
                if not self.build_from_pickle():
                    self.h = self.build_hessian()
                    self.solve()
                    self.pickle()
                    if not self.build_from_pickle():
                        raise ValueError("Could not unpickle the freshest pickle!")
            else:
                self.h = self.build_hessian()
                self.solve()
        elif solve_mol:
            self.mol.solve()

    @abc.abstractmethod
    def build_hessian(self):
        """Builds the Hessian matrix for the ENM. Implemented by subclasses.
        """
        pass

    @abc.abstractmethod
    def build_from_pickle(self):
        """Builds the solved ENM state from a pickled object. Implemented by subclasses.

        Note:
            Since ENMs in general have parameters, subclasses need to make sure to check that the input PDB and
            all parameters are identical between the pickled representation and the present run. See example in
            ENCoM class in encom module.

        Returns:
            bool: True if successful, False otherwise.
        """
        pass

    @abc.abstractmethod
    def pickle(self):
        """Uses the pickle module to produce a serialized object. Implemented by subclasses.

        Note:
            Since many ENMs may end up being run on the same file, the extension of the pickled file must be unique
            (i.e. .encom.pickle, .anm.pickle, etc.)
        """
        pass

    def solve(self):
        """Solves the Hessian matrix to get eigenvectors and eigenvalues.

        Returns:
            tuple: (eigenvalues, eigenvectors) eigenvectors are in row format
        """
        if not self.mol.solved:
            self.mol.solve()
        if self.h is None:
            self.h = self.build_hessian()
        vals, vecs = np.linalg.eigh(self.h)
        self.eigvecs_mat = vecs  # this is the version with one vector per column
        vecs = np.transpose(vecs)
        self.eigvals = vals
        self.eigvecs = vecs  # this is a list of vectors (one vector per row)
        return self.eigvals, self.eigvecs

    def is_equal(self, other):
        """Compares two ENM objects for equality. Useful for unpickling.

        Args:
            other (ENM): instance of ENM to test.

        Note:
            This is the general method that subclasses can call in addition to any specifics about their implementation.

        Returns:
            bool: True if equal, False otherwise.
        """
        assert isinstance(other, ENM)
        if not (self.pickle_version == other.pickle_version and
                        self.atypes_dict == other.atypes_dict and
                    massdefs_dicts_are_same(self.massdefs, other.massdefs) and
                    self.mol.is_equal(other.mol)):
            return False
        if len(self.mols) > 1:
            if not len(self.mols) == len(other.mols):
                return False
            for i in range(1, len(self.mols)):
                if not self.mols[i].is_equal(other.mols[i]):
                    return False
        return True

    def compute_bfactors(self, filter=None):
        """Computes the predicted b-factors.

        The root-mean-square flutuations of every mass can be seen as a prediction of the experimental temperature
        factors, or B-factors.

        Note:
            In order to have more convenient numbers, the b-factors are scaled up by a factor of 1000. This does not
            affect the validity of the value as it is only an indicator of relative flexibility between the masses.

        Args:
            filter (set, optional): set of str names of masses to select (corresponding to the names in .masses config
                                    files).

        Returns:
            list: the computed b-factors
        """
        assert self.eigvals is not None
        assert self.eigvecs is not None
        n = int(len(self.eigvecs) / 3)
        bfacts = np.zeros(n)
        if self.sq_eigvecs is None:
            self.compute_sq_eigvecs()
        for i in range(n):
            bfact = 0
            for j in range(6, len(self.eigvecs)):  # skips 1st 6 rotation-translation motions
                temp = 0
                for k in range(3):
                    temp += self.sq_eigvecs[j][3 * i + k]
                bfact += temp / self.eigvals[j]
            bfacts[i] = bfact * 1000
        if filter is not None:
            filtered_bfacts = self.filter_bfactors(bfacts, filter)
        else:
            filtered_bfacts = bfacts
        self.bfacts = filtered_bfacts
        return filtered_bfacts


    def get_exp_bfacts(self, method="average", filter=None):
        """Extracts the experimental b-factors for every mass in the system.

        Args:
            method (str): can currently be only "average" or "center". Average is the mean taken from all the atoms in
                          the residue. Center is the value from the center atom (on which the mass is positioned).
            filter (set, optional): set of str names of masses to select (corresponding to the names in .masses config
                                    files).
        """
        bfacts = []
        resis = self.mol.resis
        for m in self.mol.masses:
            resiname = m[4] # 5th field is residue key
            if filter is not None and resiname.split('|')[0].split('.')[-1] not in filter:
                continue
            resi = resis[resiname]
            if method == "average":
                s = 0
                count = 0
                for a in resi:
                    s += resi[a].bfact
                    count += 1
                bfacts.append(s / count)
            elif method == "center":
                center = m[3]  # 4th field is center atom
                bfacts.append(resi[center].bfact)
            else:
                raise ValueError('No other method than "average" or "center" currently supported by get_exp_bfacts')
        return bfacts

    def compute_vib_entropy_rigid_rotor(self):
        """Computes the vibrational entropy of the system using the rigid-rotor approximation.

        Note:
            This was the form used in the original ENCoM paper (doi: ﻿10.1371/journal.pcbi.1003569)

        Returns:
            float: The vibrational entropy of the system.
        """
        ent = 0
        for i in range(6, len(self.eigvals)):
            ent += np.log(1 / self.eigvals[i])
        self.entropy = ent
        return ent


    # def compute_entropy_oscillator(self):
    #     ent = 0
    #     for i in range(6, len(self.eigvals)):
    #         vi = (self.eigvals[i] ** 0.5) / (2 * pi)



    def filter_bfactors(self, bfacts, filter):
        if filter is not None:
            kept_bfacts = []
            masses = self.mol.masses
            for i, bfact in enumerate(self.bfacts):
                resiname = masses[i][4]
                if resiname.split('|')[0].split('.')[-1] in filter:
                    kept_bfacts.append(bfact)
            return kept_bfacts
        else:
            return bfacts

    def compute_sq_eigvecs(self):
        self.sq_eigvecs = np.square(self.eigvecs)

    def compute_bfactors_boltzmann(self, beta=1, filter=None, use_msf=False):
        assert self.eigvals is not None
        assert self.eigvecs is not None
        n = int(len(self.eigvecs) / 3)
        bfacts = np.zeros(n)
        if self.eigfreqs is None:
            self.compute_eig_freqs()
        x = self.eigfreqs * beta
        entros = x / (np.float_power(np.e, x) - 1) - np.log(1 - np.float_power(np.e, -x))
        if self.sq_eigvecs is None:
            self.compute_sq_eigvecs()
        for i in range(n):
            bfact = 0
            for j in range(6, len(self.eigvecs)):  # skips 1st 6 rotation-translation motions
                temp = 0
                for k in range(3):
                    temp += self.sq_eigvecs[j][3 * i + k]
                if use_msf:
                    bfact += temp * entros[j-6] / self.eigvals[j]
                else:
                    bfact += temp * entros[j-6]
            bfacts[i] = bfact * 1000
        if filter is not None:
            filtered_bfacts = self.filter_bfactors(bfacts, filter)
        else:
            filtered_bfacts = bfacts
        self.bfacts = filtered_bfacts
        return filtered_bfacts


    def compute_eig_freqs(self):
        nonzero = self.eigvals[6:]
        self.eigfreqs = np.sqrt(nonzero) / (2 * np.pi)


    def compute_vib_entropy(self, beta=1):
        """Vibrational entropy from the eigenfrequencies (without rigid-rotor approximation).
        Note:
            This is a more exact form of the vibrational entropy and is the preferred way to compute it. The
            compute_vib_entropy_rigid_rotor method is there only for reproducibility purposes as the old ENCoM model
            used that form.

        Args:
            beta (float, optional): The scaling factor (higher beta means lower temperature). Default is 1 due to
            preliminary data.
        Returns:
            float: The vibrational entropy of the system.
        """
        if self.teruel2021_legacy:
            beta *= 2*np.pi
        if self.eigfreqs is None:
            self.compute_eig_freqs()
        x = self.eigfreqs * beta
        return np.sum(x / (np.float_power(np.e, x) - 1) - np.log(1 - np.float_power(np.e, -x)))


    def compute_vib_entropy_vector(self, beta):
        if self.eigfreqs is None:
            self.compute_eig_freqs()
        x = self.eigfreqs * beta
        return x / (np.float_power(np.e, x) - 1) - np.log(1 - np.float_power(np.e, -x))


    def compute_Acla(self, beta=1):
        a_cla = 0
        for i in range(6, len(self.eigvals)):
            vi = np.sqrt(self.eigvals[i]) / (2 * np.pi)
            x = vi * beta
            a_cla += np.log(x)
        return 1/beta * a_cla

    def compute_Avib(self, beta=1):
        a_vib = 0
        for i in range(6, len(self.eigvals)):
            vi = np.sqrt(self.eigvals[i]) / (2 * np.pi)
            x = vi * beta
            a_vib += np.log(1/(1 - np.e ** (-1 * x)))
        return -1/beta * a_vib

    def compute_Evib(self, beta=1):
        e_vib = 0
        for i in range(6, len(self.eigvals)):
            vi = np.sqrt(self.eigvals[i]) / (2 * np.pi)
            x = vi * beta
            e_vib += x/((np.e ** x) - 1)
        return 1/beta * e_vib

    def compute_HCvib(self, beta=1):
        hc = 0
        for i in range(6, len(self.eigvals)):
            vi = np.sqrt(self.eigvals[i]) / (2 * np.pi)
            x = vi * beta
            hc += x**2 * (np.e ** x) / ((np.e ** x) - 1)**2
        return beta * hc



    def compute_vib_enthalpy(self, beta=None, factor=1):
        """Vibrational enthalpy from the eigenfrequencies (without rigid-rotor approximation).

        Args:
            beta (float, optional): The scaling factor (higher beta means lower temperature). If None, the default value
                                    is used.
            factor (float, optional): Can be used to scale the default value when beta is set to None.

        Returns:
            float: The vibrational entropy of the system.
        """
        if beta is None:
            beta = 1.55 * 10 ** -13  # beta is h/kT, this value is for T = 310 K
        beta *= factor
        e = 2.718281828459045
        h = 6.62607004 * 10 ** -34
        pi = 3.1415926535897932384626433832795028841971693993751
        enthal = 0
        for i in range(6, len(self.eigvals)):
            vi = (self.eigvals[i] ** 0.5) / (2 * pi)
            enthal += (0.5 + 1 / (e ** (beta * vi) - 1)) * h * vi
        return enthal

    def get_filtered_eigvecs_mat(self, indices=None, filter=None, transpose=True, n_vecs=None, start=0):
        """Allows to select subvectors that correspond to certain masses.

        This is useful when there are multiple masses per residue and the user wants to reduce eigenvectors to have
        1 xyz component per residue, for example. It also allows to select specific indices, which is useful in the
        case of alignments between slightly different structures.

        Note:
            The filtered vectors are always orthonormalized.

        Args:
            indices (list, optional): list of indices to select, from 0 to n-1 (n being the number of masses).
            filter (set, optional): set of str names of masses to select (corresponding to the names in .masses config
                                    files).
            transpose (bool, optional): If True, returns column vectors, otherwise returns row vectors.
            n_vecs (int, optional): The number of vectors to return, starting from `start`
            start (int, optional): The first vector to return.

        Returns:
            numpy.ndarray: the filtered eigenvectors in column format, or row format if `transpose` is False.
        """
        if n_vecs is None:
            n_vecs = len(self.eigvecs)
        if filter is None:
            filtered_vecs = self.eigvecs[:n_vecs]
        else:
            filtered_vecs = self._get_n_filtered_eigvecs(filter, n_vecs, start=0)
        if indices is not None:
            indices3n = []
            for i in indices:
                i3 = 3*i
                indices3n.append(i3)
                indices3n.append(i3+1)
                indices3n.append(i3+2)
            filtered_vecs = np.take(filtered_vecs, indices3n, axis=1)
            vecs = self._gram_schmidt(filtered_vecs, min(n_vecs, 3 * len(indices)))
        else:
            vecs = self._gram_schmidt(filtered_vecs, n_vecs)
        if transpose:
            return np.transpose(vecs)
        else:
            return vecs

    def _get_n_filtered_eigvecs(self, filter, n_vecs, start):
        masses = self.mol.masses
        n = len(masses)
        included = np.zeros((n))
        for i, m in enumerate(masses):
            if m[4].split('|')[0].split('.')[-1] in filter:
                included[i] = 1
        filtered_eigvecs = np.zeros((3*n, 3*int(np.sum(included))))
        for i in range(len(self.eigvecs)):
            eigvec = self.eigvecs[i]
            current_j = 0
            for j in range(len(eigvec)):
                if included[int(np.floor(j/3))] == 1:
                    filtered_eigvecs[i][current_j] = eigvec[j]
                    current_j += 1
        return filtered_eigvecs[start:start+n_vecs]

    def build_conf_ensemble(self, modes_list, filename, step=0.5, max_displacement=2.0, max_conformations=10000,
                            use_motion=False):
        """Creates a conformational ensemble as a multi-model PDB file.

        The idea is to make every combination of the selected modes at every given rmsd step, up to a total deformation
        for each mode (and in both directions).

        Note:
            The mode indices in the modes_list argument are assumed to start at 1, with the first nontrivial mode
            thus being at index 7.

        Args:
            modes_list (list): list of integer indices of modes to use for the conformational ensemble. The first
                               non-trivial mode is at index 7.
            filename (str): The filename where the conformational ensemble will be written.
            step (optional, float): The RMSD step between each grid point in the conformational ensemble.
            max_displacement (float, optional): The maximum RMSD displacement for each mode. Has to be a multiple of
                                                step.
            max_conformations (int, optional): The maximum number of conformations. Ensures that users do not
                                               accidentally generate huge PDB files.
        """
        assert isinstance(modes_list, list)
        grid_side = 1 + (2 * max_displacement / step)  # length of one side of the grid
        if abs((grid_side - 0.999999999999) % 2) > 0.000001:  # make sure that small floating point errors are ignored
            raise ValueError(
                "build_conf_ensemble was executed with step={0} and max_displacement={1}. max_displacement " +
                "has to be a multiple of step.".format(step, max_displacement))
        grid_side = int(round(grid_side))
        grid_side = int(grid_side)
        n_conf = int(grid_side ** len(modes_list))  # total number of points, or conformations, in the grid
        if n_conf > max_conformations:
            raise ValueError("build_conf_ensemble was executed with parameters specifying {0} conformations with " +
                             "max_conformations set at {1}. If you really want that many conformations, set " +
                             "max_conformations higher.".format(n_conf, max_conformations))
        eigvecs_list = []
        for mode_index in modes_list:
            if mode_index < 7:
                raise ValueError(
                    "build_conf_ensemble was run with mode index {0} in the modes_list, which is a trivial " +
                    "(rotational/translational) motion.".format(mode_index))
            eigvecs_list.append(np.copy(self.eigvecs[mode_index - 1]))
            modermsd = self._rmsd_of_3n_vector(eigvecs_list[-1])
            eigvecs_list[-1] /= modermsd
        with open(filename, "w") as fh:
            fh.write("REMARK Conformational ensemble written by nrgten.enm.ENM.build_conf_ensemble()\n" +
                     "REMARK from the NRGTEN Python package, copyright Najmanovich Research Group 2020\n" +
                     "REMARK This ensemble contains {0} conformations\n".format(n_conf))
            conf_seq = range(n_conf)
            if use_motion:
                if len(modes_list) > 1:
                    raise ValueError("Cannot use 'motion' style conformational ensemble with more than one mode")
                conf_seq = []
                for breaks in [[int((n_conf-1)/2), n_conf, 1], [n_conf-1, int((n_conf-1)/2)-1, -1],
                               [int((n_conf-1)/2)-1, -1, -1], [0, int((n_conf-1)/2), 1]]:
                    for i in range(*breaks):
                        conf_seq.append(i)
            for i in conf_seq:
                self._write_one_point(eigvecs_list, i, grid_side, step, fh)
            fh.write("END\n")

    def build_motion_elastic(self, mode_index, filename, max_displacement=2.0, n_conf=90):
        if mode_index < 7:
            raise ValueError("build_motion_elastic() called with trivial mode (mode # {})".format(mode_index))
        vec = np.copy(self.eigvecs[mode_index-1])
        modermsd = self._rmsd_of_3n_vector(vec)
        vec /= modermsd
        full_path = []
        for i in range(n_conf):
            angle = np.pi*2*i/float(n_conf)
            full_path.append(np.sin(angle)*max_displacement)
        with open(filename, "w") as fh:
            fh.write("REMARK Elastic motion for mode {} written by " +
                     "nrgten.enm.ENM.build_elastic_motion()\n".format(mode_index) +
                     "REMARK from the NRGTEN Python package, copyright Najmanovich Research Group 2020\n" +
                     "REMARK This ensemble contains {0} conformations\n".format(n_conf))
            for i, rmsd in enumerate(full_path):
                t_vect = vec * rmsd
                self._translate_3n_vector(t_vect)
                self._write_model(i+1, fh)
                self._translate_3n_vector(-t_vect)


    def build_conf_ensemble_with_nrg(self, modes_list, filename_ensemble, filename_energies, base_energy, beta,
                                     eigval_scaling=0.1, step=0.5, max_displacement=2.0, max_conformations=10000,
                                     use_motion=False):
        assert isinstance(modes_list, list)
        eigvals_list = np.array([self.eigvals[x-1]*eigval_scaling for x in modes_list])

        grid_side = 1 + (2 * max_displacement / step)  # length of one side of the grid
        if abs((grid_side - 0.999999999999) % 2) > 0.000001:  # make sure that small floating point errors are ignored
            raise ValueError(
                "build_conf_ensemble was executed with step={0} and max_displacement={1}. max_displacement " +
                "has to be a multiple of step.".format(step, max_displacement))
        grid_side = int(round(grid_side))
        grid_side = int(grid_side)
        n_conf = int(grid_side ** len(modes_list))  # total number of points, or conformations, in the grid
        if n_conf > max_conformations:
            raise ValueError("build_conf_ensemble was executed with parameters specifying {0} conformations with " +
                             "max_conformations set at {1}. If you really want that many conformations, set " +
                             "max_conformations higher.".format(n_conf, max_conformations))
        eigvecs_list = []
        for mode_index in modes_list:
            if mode_index < 7:
                raise ValueError(
                    "build_conf_ensemble was run with mode index {0} in the modes_list, which is a trivial " +
                    "(rotational/translational) motion.".format(mode_index))
            eigvecs_list.append(np.copy(self.eigvecs[mode_index - 1]))
            modermsd = self._rmsd_of_3n_vector(eigvecs_list[-1])
            eigvecs_list[-1] /= modermsd
        energies = np.zeros(n_conf)
        print(eigvals_list)
        with open(filename_ensemble, "w") as fh:
            fh.write("REMARK Conformational ensemble written by nrgten.enm.ENM.build_conf_ensemble()\n" +
                     "REMARK from the NRGTEN Python package, copyright Najmanovich Research Group 2020\n" +
                     "REMARK This ensemble contains {0} conformations\n".format(n_conf))
            conf_seq = range(n_conf)
            if use_motion:
                if len(modes_list) > 1:
                    raise ValueError("Cannot use 'motion' style conformational ensemble with more than one mode")
                conf_seq = []
                for breaks in [[int((n_conf - 1) / 2), n_conf, 1], [n_conf - 1, int((n_conf - 1) / 2) - 1, -1],
                               [int((n_conf - 1) / 2) - 1, -1, -1], [0, int((n_conf - 1) / 2), 1]]:
                    for i in range(*breaks):
                        conf_seq.append(i)
            for i in conf_seq:
                energies[i] = self._write_one_point_nrg(eigvecs_list, eigvals_list, i, grid_side, step, fh)
            fh.write("END\n")
        energies += base_energy
        probabilities = np.e**(-beta*energies)
        probabilities /= np.sum(probabilities)
        with open(filename_energies, "w") as f:
            f.write("model energy beta probability\n")
            for i in range(len(energies)):
                f.write("{} {} {} {}\n".format(i+1, energies[i], beta, probabilities[i]))


    def _write_one_point(self, eigvecs_list, conf_n, grid_side, step, fh):
        """Helper function for build_conf_ensemble. conf_n is the conformation number from 0 to n-1.

        Computes contributions from every mode automatically, translates the enm coordinates, and resets it back to
        original after having written one model to the filehandle fh.
        """
        model_n = conf_n
        nsteps_list = []
        for i in range(len(eigvecs_list)):
            nsteps = conf_n % grid_side
            conf_n -= nsteps
            conf_n /= grid_side
            nsteps_list.append(nsteps - (grid_side - 1) / 2)
        t_vect = np.zeros(len(eigvecs_list[0]))
        for vec, nsteps in zip(eigvecs_list, nsteps_list):
            t_vect += vec * nsteps * step
        self._translate_3n_vector(t_vect)
        self._write_model(int(model_n + 1), fh)
        self._translate_3n_vector(-t_vect)

    def _write_one_point_nrg(self, eigvecs_list, spring_constants, conf_n, grid_side, step, fh):
        """Helper function for build_conf_ensemble. conf_n is the conformation number from 0 to n-1.

        Computes contributions from every mode automatically, translates the enm coordinates, and resets it back to
        original after having written one model to the filehandle fh.
        """
        model_n = conf_n
        nsteps_list = []
        for i in range(len(eigvecs_list)):
            nsteps = conf_n % grid_side
            conf_n -= nsteps
            conf_n /= grid_side
            nsteps_list.append(nsteps - (grid_side - 1) / 2)
        t_vect = np.zeros(len(eigvecs_list[0]))
        nrg = 0
        for vec, nsteps, k in zip(eigvecs_list, nsteps_list, spring_constants):
            t_vect += vec * nsteps * step
            nrg += k * (nsteps*step)**2
        self._translate_3n_vector(t_vect)
        self._write_model(int(model_n + 1), fh)
        self._translate_3n_vector(-t_vect)
        return nrg


    def _rmsd_of_3n_vector(self, vec):
        dists = np.zeros((int(len(vec) / 3)))
        for i in range(0, len(vec), 3):
            dists[int(i / 3)] = vec[i] ** 2 + vec[i + 1] ** 2 + vec[i + 2] ** 2
        return np.sqrt(np.mean(dists))

    def _write_model(self, count, fh):
        """ Writes one model to the given filehandle, with model number count.
        """
        fh.write("MODEL     {:>4}\n".format(count))
        self._write_to_filehandle(fh)
        fh.write("ENDMDL\n")

    def _clear_info(self):
        """Clears the Hessian and row eigenvectors before pickling, to save space.
        """
        self.h, self.eigvecs = None, None

    def _reconstitute(self):
        """Reconstitues the row eigenvectors from the column eigenvectors when building from pickle.
        """
        self.eigvecs = np.transpose(self.eigvecs_mat)

    def _parse_massdefs(self, atypes_list, massdef_list):
        massdefs = dict()
        for i, (atypes_file, massdef_file) in enumerate(zip(atypes_list, massdef_list)):
            assert atypes_file.split('/')[-1].split('.')[0] == massdef_file.split('/')[-1].split('.')[0]
            resis = self._resis_from_atypes(atypes_file)
            massdef = MassDef(massdef_file)
            for r in resis:
                if r in massdefs:
                    raise ValueError(
                        "Trying to add residue {0} twice in a mass definition!".format(r))
                massdefs[r] = massdef
        return massdefs

    def _resis_from_atypes(self, atypes):
        resis = []
        with open(atypes) as f:
            lines = f.readlines()
        for line in lines:
            resi = line.split('|')[0].strip()
            if resi != "ADD_TO_ABOVE":
                resis.append(resi)
        return resis

    def _build_atypes_dict(self, atypes_list):
        atd = dict()
        for filename in atypes_list:
            d = self._get_atype_dict(filename)
            for key in d:
                if key in atd:
                    raise ValueError(
                        "Residue {0} was in many atomypes files (2nd of which is {1}".format(key, filename))
                atd[key] = d[key]
        return atd

    def _get_atype_dict(self, filename):
        d = dict()
        with open(filename) as f:
            lines = f.readlines()
        for l in lines:
            ll = l.split('|')
            resi = ll[0].strip()
            rest = ll[1].strip()
            # atoms is a list of 2-element lists : [atom name, atom type number]
            atoms = [a.strip().split(':') for a in rest.split(',')]
            if len(resi) > 3 and resi == "ADD_TO_ABOVE":  # line with atom types to add to many residues
                for atom in atoms:
                    for key in d:
                        if atom[0] in d[key]:
                            raise ValueError("Trying to add atom {0} twice (file {1})".format(atom[0], filename))
                        d[key][atom[0]] = int(atom[1])
            elif len(resi) > 3:  # resiname is too long for PDB files
                raise ValueError(
                    "unsupported (longer than 3 characters) residue name {0} in file {1}".format(resi, filename))
            else:  # normal line, one residue
                if resi in d:
                    raise ValueError("Trying to add residue {0} 2 times (file {1})".format(resi, filename))
                d[resi] = dict()
                for atom in atoms:
                    if atom[0] in d[resi]:
                        raise ValueError(
                            "Trying to add atom {0} 2 times to residue {1} (file {2})".format(atom[0], resi, filename))
                    d[resi][atom[0]] = int(atom[1])
        return d

    def _print_verbose(self, string):
        if self.verbose:
            print(string)

    def get_n_masses(self):
        """Gives the number of masses in the ENM.

        Returns:
            int: the number of masses in the ENM.
        """
        return len(self.mol.masses)

    def get_xyz(self):
        """Gives the x,y,z coordinates of the masses in the form of an Nx3 matrix.

        Returns:
            numpy.ndarray: Nx3 matrix of x,y,z coordinates of the masses.
        """
        return self.mol.get_xyz()

    def get_filtered_xyz(self, filter):
        """Gives the x,y,z coordinates of selected masses in the form of an Nx3 matrix.

        Args:
            filter (set): set of str names of masses to select (corresponding to the names in .masses config
                          files).
        Returns:
            numpy.ndarray: Nx3 matrix of x,y,z coordinates of the selected masses.
        """
        return self.mol.get_filtered_xyz(filter)

    def get_filtered_3n_vector(self, filter):
        """Gives the x,y,z coordinates of selected masses in the form of a single vector.

        Note:
            The vector is in the format: [x1, y1, z1, x2, y2, z2, ..., xN, yN, zN].

        Args:
            filter (set): set of str names of masses to select (corresponding to the names in .masses config
                          files).
        Returns:
            The 3N-dimensional vector of coordinates for the selected masses.
        """
        return self.mol.get_filtered_3n_vector(filter)

    def set_bfactors(self, vector):
        """Sets the b-factors of all atoms in the file.

        Args:
            vector (list): the new b-factors, needs to be the same length as the number of masses.
        """
        assert self.get_n_masses() == len(vector)
        self.mol.set_bfactors(vector)

    def set_occupancy(self, vector):
        """Sets the occupancy of all atoms in the file.

        Args:
            vector (list): the new occupancy values, need to be the same length as the number of masses.
        """
        assert self.get_n_masses() == len(vector)
        self.mol.set_occupancy(vector)

    def write_dynamical_signature(self, outfile, normalize=False):
        """Writes a 'dynamical signature' to the output file specified.

        The dynamical signature is simply the predicted b_factor at every mass.

        Args:
            outfile (str): the output file.
            normalize (bool, optional): if True, the signature is normalized to 1
        """
        if not hasattr(self, "bfacts") or self.bfacts is None:
            self.compute_bfactors()
        labels = self.get_mass_labels()
        bfacts = self.bfacts
        divider = 1
        if normalize:
            divider = np.sum(bfacts)
        assert len(labels) == len(bfacts)
        with open(outfile, "w") as f:
            for label, bfact in zip(labels, bfacts):
                f.write("{0}\t{1}\n".format(label, bfact / divider))

    def get_mass_labels(self, use_xyz=False):
        masses = self.mol.masses
        labels = []
        for mass in masses:
            if use_xyz:
                for axis in ['X', 'Y', 'Z']:
                    labels.append("{}|{}".format(mass[4], axis))
            else:
                labels.append(mass[4])
        return labels

    def _write_to_file(self, filename):
        self.mol.write_to_file(filename)

    def _write_to_filehandle(self, fh):
        self.mol.write_to_filehandle(fh)

    def _translate_3n_matrix(self, vector):
        self.mol.translate_3n_matrix(vector)
        self.mol.update()

    def _translate_3n_vector(self, vector):
        self.mol.translate_3n_vector(vector)
        self.mol.update()

    def _translate_xyz(self, vector):
        self.mol.translate_xyz(vector)
        self.mol.update()

    def _rotate(self, rotation_mat):
        self.mol.rotate(rotation_mat)
        self.mol.update()

    def _gram_schmidt(self, vectors, target_n):
        """ From https://gist.github.com/iizukak/1287876/edad3c337844fac34f7e56ec09f9cb27d4907cc7#gistcomment-1871542.
            Gram-schmidt orthonormalization of row vectors.
        """
        basis = []
        for v in vectors:
            x = np.zeros((len(v)))
            for b in basis:
                x += np.dot(v, b)*b
            w = v - x
            # w = v - np.sum(np.dot(v,b)*b  for b in basis) # Calling np.sum(generator) is deprecated
            if (w > 1e-10).any():
                basis.append(w/np.linalg.norm(w))
            if len(basis) == target_n:
                break
        if len(basis) < target_n:
            raise ValueError("Unable to produce {0} orthonormalized vectors for pdb file {1}".format(target_n, self.pdb_file))
        return np.array(basis)

    def _get_weighted_surf_atomnum(self, sd, num_a, num_b, resi_a, resi_b, anums_dict):
        try:
            i = self.atypes_dict[resi_a][anums_dict[num_a].name]
        except KeyError:
            self.not_there.add((resi_a, anums_dict[num_a]))
            return 0
        try:
            j = self.atypes_dict[resi_b][anums_dict[num_b].name]
        except KeyError:
            self.not_there.add((resi_b, anums_dict[num_b]))
            return 0
        return sd[num_a][num_b] * self.inter_mat[i][j]

    def _get_total_surf_atomnum(self, sd, anums_list, i, j, resi_a, resi_b, anums_dict):
        if resi_a == "LIG" and resi_b == "ASP":
            dum = 0
        surf_tot = 0
        for num_a in anums_list[i]:
            for num_b in anums_list[j]:
                try:
                    if num_b in sd[num_a]:
                        surf_tot += self._get_weighted_surf_atomnum(sd, num_a, num_b, resi_a, resi_b, anums_dict)
                    if num_a in sd[num_b]:
                        surf_tot += self._get_weighted_surf_atomnum(sd, num_b, num_a, resi_b, resi_a, anums_dict)
                except:
                    print("Error in surface computation, residues {0} and {1}, file {2}".format(resi_a, resi_b,
                                                                                                self.mol.pdb_file))
                    raise
        return surf_tot

    def _compute_Bij(self, masses, resis):
        """ Computes the Bij term in the ENCoM potential, which is the
            modulating factor of the long-range interaction (V4) term calculated
            according to atomic surface complementarity.
        """
        sd = self.mol.get_surface_dict()
        n = len(masses)
        bij = np.zeros((n, n))
        atomnums_list = self._make_atomnums_list(masses, resis)
        atomnums_dict = self._make_atomnums_dict(resis)
        self._print_verbose(self.atypes_dict)
        for i in range(n):
            ma = masses[i]  # mass_a
            ka = ma[4]  # key_a
            resiname_a = ka.split('|')[0].split('.')[0]
            for j in range(i + 1, n):
                mb = masses[j]
                kb = mb[4]
                resiname_b = kb.split('|')[0].split('.')[0]
                try:
                    surf_tot = self._get_total_surf_atomnum(sd, atomnums_list, i, j, resiname_a, resiname_b, atomnums_dict)
                except:
                    e = sys.exc_info()[0]
                    print("Error in surface computation, file {0}".format(self.mol.pdb_file))
                    raise e
                if surf_tot < 0:
                    surf_tot = 0
                bij[i][j] = surf_tot
                bij[j][i] = surf_tot
        self._print_verbose(self.not_there)
        return bij

    def _make_atomnums_list(self, masses, resis):
        """ Makes a list which corresponds in index with the masses received and
            which contains lists of atom numbers for every mass.
        """
        assert len(masses) == len(resis)
        n = len(masses)
        atomnums_list = []
        for i in range(n):
            resi = resis[masses[i][4]]
            anums = []
            for a in resi:
                anums.append(resi[a].num)
            atomnums_list.append(anums)
        return atomnums_list

    def _make_atomnums_dict(self, resis):
        """ Makes a dictionary associating each atom number to an Atom object.
        """
        numdict = dict()
        for r in resis:
            for a in resis[r]:
                numdict[resis[r][a].num] = resis[r][a]
        return numdict



# End of ENM class, start of utility methods ###########################################################################


def generate_dynasigs_df(filenames, outname, id_func=None, beta_values=None, models=None, models_labels=None,
                         additional_info_dict=None, add_info_labels=None, added_atypes_list=None,
                         added_massdef_list=None, get_cf_from_file=False, use_xyz=False, use_msf=False):
    if models is None:
        from nrgten.encom import ENCoM
        models = [ENCoM]
        models_labels = ["ENCoM"]
    if beta_values is None:
        beta_values = [1]
    if id_func is None:
        id_func = lambda x: x.split('/')[-1]
    assert isinstance(beta_values, list) and isinstance(models, list) and isinstance(models_labels, list)
    assert len(models) == len(models_labels)
    if added_atypes_list is not None:
        assert len(added_atypes_list) == len(filenames) == len(added_massdef_list)
    dynasigs_data = []
    dynasigs_masslabels = None
    ids_data = []
    betas_data = []
    vib_entro_data = []
    mod_labels_data = []
    cfs = []
    for mod, mod_lab in zip(models, models_labels):
        for i, filename in enumerate(filenames):
            filename_id = id_func(filename)
            if added_atypes_list is None:
                enm = mod(filename)
            else:
                enm = mod(filename, added_atypes=added_atypes_list[i], added_massdef=added_massdef_list[i])
            masslabels = [x.split('.')[-1] for x in enm.get_mass_labels(use_xyz=use_xyz)]  # only the mass name to allow for mutations
            for beta_val in beta_values:
                dynasigs_data.append(enm.compute_bfactors_boltzmann(beta=beta_val, use_xyz=use_xyz, use_msf=use_msf))
                vib_entro_data.append(enm.compute_vib_entropy(beta=beta_val))
                if dynasigs_masslabels is None:
                    dynasigs_masslabels = masslabels
                else:
                    validate_masslabels(masslabels, dynasigs_masslabels)
                ids_data.append(filename_id)
                betas_data.append(beta_val)
                mod_labels_data.append(mod_lab)
                if get_cf_from_file:
                    cf = None
                    with open(filename) as f:
                        lines = f.readlines()
                    for line in lines:
                        if line.startswith("REMARK CF="):
                            cf = float(line[10:])
                            break
                    if cf is not None:
                        cfs.append(cf)
                    else:
                        raise ValueError("Could not find CF line in file {}".format(filename))
    colnames = ["id", "model", "beta", "vib_entro"]
    if get_cf_from_file:
        colnames.append("cf")
    if additional_info_dict is not None:
        assert add_info_labels is not None and isinstance(add_info_labels, list)
        colnames = colnames + add_info_labels
    sig_len = len(dynasigs_masslabels)
    colnames = colnames + dynasigs_masslabels
    df_strs = [" ".join(colnames)]
    for i in range(len(dynasigs_data)):
        assert len(dynasigs_data[i]) == sig_len
        current_str = "{} {} {} {}".format(ids_data[i], mod_labels_data[i], betas_data[i], vib_entro_data[i])
        if get_cf_from_file:
            current_str = current_str + " {}".format(cfs[i])
        if additional_info_dict is not None:
            current_str = current_str + " " + " ".join(
                [str(additional_info_dict[ids_data[i]][label]) for label in add_info_labels])
        current_str = current_str + " " + " ".join([str(x) for x in dynasigs_data[i]])
        df_strs.append(current_str)
    with open(outname, "w") as f:
        f.write("\n".join(df_strs))
        f.write("\n")


def validate_masslabels(new_labels, old_labels):
    assert len(new_labels) == len(old_labels)
    for new, old in zip(new_labels, old_labels):
        if new != old:
            raise ValueError("problem with masslabels in generate_dynasigs_df: {} not equal to {}".format(new, old))





import numpy as np

class Atom:
    __slots__ = ['num', 'name', 'alt', 'resiname', 'chain', 'resinum', 'insert',
                 'xyz', 'occ', 'bfact', 'seg_id', 'elem', 'charge', 'parent_resi', 'macromol_index']

    def __init__(self, line):
        n = len(line)
        if n < 54:
            raise ValueError("Trying to create Atom with line shorter than 54")
        self.parent_resi = None
        self.macromol_index = None
        self.seg_id = None
        self.num = int(line[6:11])
        self.name = line[12:16].strip()
        self.alt = line[16]
        self.resiname = line[17:20].strip()
        self.chain = line[21]
        self.resinum = int(line[22:26])
        self.insert = line[26]
        self.xyz = np.array([float(line[30:38]),  float(line[38:46]), float(line[46:54])])
        # placeholders for short lines
        self.occ = 0
        self.bfact = 0
        self.elem = "  "
        self.charge = "  "
        try:
            if n >= 80:
                self.charge = line[78:80]
                self.elem = line[76:78]
                self.bfact = self.get_float(line[60:66])
                self.occ = self.get_float(line[54:60])
            elif n >= 78:
                self.elem = line[76:78]
                self.bfact = self.get_float(line[60:66])
                self.occ = self.get_float(line[54:60])
            elif n >= 66:
                self.bfact = self.get_float(line[60:66])
                self.occ = self.get_float(line[54:60])
            elif n >= 60:
                self.occ = self.get_float(line[54:60])
        except ValueError as e:
            raise

    def get_float(self, s):
        """ Returns 0 for empty strings, float(s) otherwise.
        """
        if len(s.strip()) == 0:
            return 0.0
        return float(s)


    def __getitem__(self, i):
        if i == 0:
            return self.x
        elif i == 1:
            return self.y
        elif i == 2:
            return self.z
        else:
            raise ValueError("Trying to index Atom object at index other than 0, 1 or 2 (x, y, z)")

    def translate(self, vec):
        assert len(vec) == 3
        self.xyz += np.array(vec)

    def set_xyz(self, xyz):
        self.xyz = np.array([xyz[0], xyz[1], xyz[2]])

    def set_parent_resi(self, resi):
        self.parent_resi = resi

    def set_macromol_index(self, i):
        self.macromol_index = i

    def is_equal(self, other):
        assert isinstance(other, Atom)
        a=self.num == other.num
        b=self.name == other.name
        c=self.alt == other.alt
        d=self.resiname == other.resiname
        e=self.chain == other.chain
        f=self.resinum == other.resinum
        g=self.insert == other.insert
        h=np.array_equal(self.xyz, other.xyz)
        i=self.occ == other.occ
        j=self.bfact == other.bfact
        k=self.seg_id == other.seg_id
        l=self.elem == other.elem
        m=self.charge == other.charge
        n=self.parent_resi == other.parent_resi
        o=self.macromol_index == other.macromol_index
        if not (self.num == other.num and
                self.name == other.name and
                self.alt == other.alt and
                self.resiname == other.resiname and
                self.chain == other.chain and
                self.resinum == other.resinum and
                self.insert == other.insert and
                np.array_equal(self.xyz, other.xyz) and
                self.occ == other.occ and
                self.bfact == other.bfact and
                self.seg_id == other.seg_id and
                self.elem == other.elem and
                self.charge == other.charge and
                self.parent_resi == other.parent_resi and
                self.macromol_index == other.macromol_index):
            return False
        return True

    def __str__(self):
        return self.toString()

    def __repr__(self):
        return self.toString()

    def toString(self):
        x, y, z = self.xyz
        spaces10 = "          "
        if len(self.name) <= 3:
            return "ATOM  {:>5}  {:3}{}{:>3} {}{:>4}    {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}{}{}{}".format(
                self.num, self.name, self.alt, self.resiname, self.chain, self.resinum, x, y, z,
                self.occ, self.bfact, spaces10, self.elem, self.charge)
        # difference is the added space for atom name (weird "center-alignment" in PDB)
        return "ATOM  {:>5} {:4}{}{:>3} {}{:>4}    {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}{}{}{}".format(
            self.num, self.name, self.alt, self.resiname, self.chain, self.resinum, x, y, z,
            self.occ, self.bfact, spaces10, self.elem, self.charge)



digits = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"}


class ResidueList:
    """ Class used to represent the list of all kept residues from a macromol
    file (PDB or other), as a dict of dicts. This version keeps a single list
    for every residue type but needs a dict of atom types."""

    def __init__(self, atypes_dict):
        self.list = {}
        self.alt_dict = {}
        self.atypes_dict = atypes_dict
        self.lastkey = None
        self.resicount = 0
        self.atoms = {}  # stores atoms according to their atom number and validates there is no repeat
        self.alt_flag = False

    def add(self, atom):
        if atom.name[0] == 'H':  # Hydrogen atoms are not considered
            return
        elif atom.name[0] in digits and atom.name[1] == "H":  # Also hydrogen atom
            return
        elif atom.elem == 'H':  # Hydrogen
            return
        if self.atypes_dict is None:
            self.add_pvt(atom)
        elif atom.resiname in self.atypes_dict:
            self.add_pvt(atom)
        else:
            raise ValueError("Trying to add atom with unsupported residue type: \n{0}\n".format(atom.toString()))

    def add_pvt(self, atom):
        if atom.insert == ' ':
            basekey = "{}|{}|{}".format(atom.resiname, atom.resinum, atom.chain)
        else:
            basekey = "{}|{}{}|{}".format(atom.resiname, atom.resinum, atom.insert, atom.chain)
        if atom.alt != ' ':  # if alternate locations, first one seen has precedence, others are ignored
            if basekey in self.alt_dict:
                if atom.alt != self.alt_dict[basekey]:
                    return
            else:
                self.alt_dict[basekey] = atom.alt
        if basekey != self.lastkey:
            self.resicount += 1
        self.lastkey = basekey
        key = basekey + "|" + str(self.resicount)
        if key not in self.list:
            self.list[key] = {}
        d = self.list[key]
        if atom.name in d:  # treats cases with alternate locations at the residue level
            if atom.alt != ' ':
                self.alt_flag = True
                return
            print(d)
            print(atom.name)
            print(key)
            raise ValueError("Trying to add same atom twice!")
        if atom.num in self.atoms:
            raise ValueError("Atom number {0} found twice in pdb file!".format(atom.num))
        self.atoms[atom.num] = atom
        d[atom.name] = atom

    def get_resilist(self):
        return self.list

    def get_ordered_atoms(self):
        return [self.atoms[k] for k in sorted(self.atoms.keys())]
